# (C)2018 Ingram Micro Inc. all Rights Reserve
import logging

logger = logging.getLogger()


class CleanUpProcedureBase(object):
    def __init__(self, name, list_of_available_steps):
        self.list_of_available_steps = list_of_available_steps
        self.name = name

    def run(self, clean_up_step, check_only, **args):
        try:
            func, checks = [(x.function, x.checks) for x in self.list_of_available_steps if x.name == clean_up_step][0]
        except:
            raise NotImplementedError(
                'The cleanup step "%s" is not available for the %s module.' % (clean_up_step, self.name)
            )

        is_all_checkers_passes = True
        checkers_issues = []

        if len(checks) > 0:
            logger.info('Starting checks for the cleanup step "%s"...' % clean_up_step)
            for check in checks:
                if hasattr(self, check):
                    check_result, issues = getattr(self, check)(**args)
                    checkers_issues.extend(issues)
                    is_all_checkers_passes &= check_result
                else:
                    raise NotImplementedError(
                        'The check "%s" is not implemented for the %s module.' % (clean_up_step, self.name)
                    )

            if not is_all_checkers_passes:
                logger.error('Some of the checks could not be passed. Resolve all issues and perform this step again.')
                for issue in checkers_issues:
                    logger.error(issue)

                return 1
            else:
                logger.info('All checks for the cleanup step "%s" were passed successfully.' % clean_up_step)
        else:
            logger.debug('There are no checks for the cleanup step "%s".' % clean_up_step)

        if check_only:
            logger.info("Now you can perform this step by running this command without the '--check-only' argument.")
            return 0

        if hasattr(self, func):
            getattr(self, func)(**args)
        else:
            raise NotImplementedError(
                'The cleanup step %s is not implemented for the %s module.' % (clean_up_step, self.name)
            )

        return 0
