# (C)2021 Ingram Micro Inc. all Rights Reserved
from poaupdater import uBilling, pbaapi

from oacleanup.cleanup.base import CleanUpProcedureBase
from oacleanup.settings import Step, formatter, BillingModule
from oacleanup.utils.billing import is_billing_installed
from oacleanup.utils.billing_db import get_connection_to_billing_db, is_billing_db_table_exists
from oacleanup.utils.errors import CleanUpException
from oacleanup.utils.hosts import file_exists, remove_files
from psycopg2.extras import RealDictCursor
from xmlrpclib import Fault

import csv
import logging
import os
import time
import traceback

logger = logging.getLogger()

DRY_RUN = False
HELP_LINK = 'https://cloudblue.freshdesk.com/a/solutions/articles/44002195657'
DECRYPTED_CERTS_DATA_FILENAME_PREFIX = 'decrypted_certs_data_'
CERTGATE_ID = 1115

DEPRECATED_DB_TABLES_REGISTRAR_CERT_INFO = ['CertEnomCert', 'OpenSRSCert', 'GlobalSignCert', 'SSLStoreCert']

OBSOLETE_FILES = [
    '/usr/local/bm/conf/wnd/BM/customization.CERTGATE',
    '/usr/local/bm/etc/ssm.conf.d/.CERTENOM.conf',
    '/usr/local/bm/etc/ssm.conf.d/.CERTOPENSRS.conf',
    '/usr/local/bm/etc/ssm.conf.d/.CERTGLOBALSIGN.conf',
    '/usr/local/bm/etc/ssm.conf.d/.CERTSSLSTORE.conf',
    '/usr/local/bm/etc/ssm.conf.d/.CERTGATE.conf'
]

DEPRECATED_CERT_CONTAINERS = ['CERTGATE', 'CERTENOM', 'CERTOPENSRS', 'CERTSSLSTORE', 'CERTGLOBALSIGN']

QUERY_COUNT_CERT_PLUGIN_TEMPLATES = 'SELECT COUNT(*) FROM "CertPluginTemplate"'
QUERY_COUNT_CERT_DB_CONTAINERS = 'SELECT COUNT(*) FROM "DBContainers" WHERE "Name" IN (\'%s\')' % \
    "','".join(map(lambda p: p + " " * (80 - len(p)), DEPRECATED_CERT_CONTAINERS))
QUERY_COUNT_CERT_REGISTERED_SERVER = 'SELECT COUNT(*) FROM "RegisteredServer" WHERE "SrvID" = %d AND ' \
    '"ContName" = \'CERTGATE\'' % CERTGATE_ID
QUERY_COUNT_CERT_SERVICE_GATE = 'SELECT COUNT(*) FROM "ServiceGate" WHERE "GateID" = %d AND "ContName" = \'CERTGATE\'' \
    % CERTGATE_ID
QUERY_COUNT_CERT_METHODS = 'SELECT COUNT(*) FROM "Method" WHERE "Container" IN (\'%s\')' % \
    "','".join(DEPRECATED_CERT_CONTAINERS)
QUERY_COUNT_CERT_VENDOR_DEFAULT_CONFIG = 'SELECT COUNT(*) FROM "VendorDefaultConfig" WHERE "Container" = \'CERTGATE\''


SHOW_OBJECTS_STEP = Step(
    name='show-objects',
    description=formatter(
        'Show all objects related to the deprecated {module} which must be removed before upgrading CloudBlue from '
        '20.5 to 21'
    ),
    function='check_certgate_is_removed',
    checks=[],
    arguments={}
)

EXPORT_CERT_INFO_STEP = Step(
    name='export-cert-data',
    description=formatter(
        'Export the information about C++ SSL certificates in CSV format. ATTENTION: the file will contain sensitive '
        'data and should be secured.'
    ),
    function='export_certs',
    checks=[],
    arguments={}
)

CLEAN_CERTIFICATES_STEP = Step(
    name='clean-certificates',
    description=formatter('Clean {module} and archive the related subscriptions and orders.'),
    function='clean_certs_subs_orders',
    checks=[],
    arguments={}
)

CLEAN_PLANS_STEP = Step(
    name='clean-service-plans',
    description=formatter('Archive Documents, Service Plans and Service Templates related to {module}.'),
    function='archive_docs_plans_templates',
    checks=[],
    arguments={}
)

CLEAN_CERTGATE_STEP = Step(
    name='certgate',
    description=formatter('Delete CERTGATE files and clean CERTGATE data from Billing DB.'),
    function='clean_certgate',
    checks=[],
    arguments={}
)


class CertSubObject():
    def __init__(self, sub_name, sub_id, cert_name, cert_id, customer_name, customer_id):
        self.sub_name = sub_name
        self.sub_id = sub_id
        self.cert_name = cert_name
        self.cert_id = cert_id
        self.customer_name = customer_name
        self.customer_id = customer_id

    def get_cert_info(self):
        return 'Subscription ID #%s ("%s") Certificate "%s" (ID #%s) Customer "%s" (ID #%s)' % (
            str(self.sub_id), self.sub_name, self.cert_name, str(self.cert_id), self.customer_name, str(self.customer_id)
        )


class VendorObject():
    def __init__(self, vendor_name):
        self.name = vendor_name
        self.subs = list()

    def add_cert_obj(self, cert_sub_obj):
        self.subs.append(cert_sub_obj)


class CleanUpCppCertGate(CleanUpProcedureBase):
    def __init__(self):
        super(CleanUpCppCertGate, self).__init__(
            name=BillingModule.CPP_CERT_GATE,
            list_of_available_steps=[
                SHOW_OBJECTS_STEP, EXPORT_CERT_INFO_STEP, CLEAN_CERTIFICATES_STEP, CLEAN_PLANS_STEP, CLEAN_CERTGATE_STEP
            ],
        )
        self.help_link = HELP_LINK
        self.billing_is_installed = is_billing_installed()
        if self.billing_is_installed:
            self.pba_db_connection = get_connection_to_billing_db()

    def check_certgate_is_removed(self):
        objects_found = False

        if self._find_and_report_cert_subs_info():
            objects_found = True

        if self._find_and_report_cert_plans_info():
            objects_found = True

        if self._find_and_report_cert_templates_info():
            objects_found = True

        if self._find_and_report_cert_plugins_info():
            objects_found = True

        if self._find_and_report_certgate_rows_bss_db(QUERY_COUNT_CERT_PLUGIN_TEMPLATES, 'CertPluginTemplate', True):
            objects_found = True

        if self._find_and_report_certgate_rows_bss_db(QUERY_COUNT_CERT_DB_CONTAINERS, 'DBContainers'):
            objects_found = True

        if self._find_and_report_certgate_rows_bss_db(QUERY_COUNT_CERT_REGISTERED_SERVER, 'RegisteredServer'):
            objects_found = True

        if self._find_and_report_certgate_rows_bss_db(QUERY_COUNT_CERT_SERVICE_GATE, 'ServiceGate'):
            objects_found = True

        if self._find_and_report_certgate_rows_bss_db(QUERY_COUNT_CERT_METHODS, 'Method'):
            objects_found = True

        if self._find_and_report_certgate_rows_bss_db(QUERY_COUNT_CERT_VENDOR_DEFAULT_CONFIG, 'VendorDefaultConfig'):
            objects_found = True

        if self._find_and_report_cert_files():
            objects_found = True

        if not objects_found:
            logger.info("Information related to deprecated CERTGATE not found")

    # Return True if certificate subscriptions found
    def _find_and_report_cert_subs_info(self):
        check_db_tables = ['Cert', 'CertST']
        for check_db_table in check_db_tables:
            if not is_billing_db_table_exists(self.pba_db_connection, check_db_table):
                logger.debug("Certificate subscriptions cannot exist: the BSS DB table '%s' not found" % check_db_table)
                return False

        result = dict()
        with self.pba_db_connection.cursor(cursor_factory=RealDictCursor) as curs:
            curs.execute(
                'SELECT c."CertID", '
                'c."subscriptionID", '
                'c."Name" as "certName", '
                's."AccountID", '
                's."SubscriptionName", '
                'a."CompanyName", '
                'a."VendorAccountID", '
                'va."CompanyName" as "VendorName" '
                'FROM "Cert" as c '
                'JOIN "Subscription" as s ON c."subscriptionID" = s."subscriptionID" '
                'JOIN "Account" as a ON s."AccountID" = a."AccountID" '
                'JOIN "CertST" as cst ON s."serviceTemplateID" = cst."serviceTemplateID" '
                'JOIN "Account" as va ON a."VendorAccountID" = va."AccountID" '
                'ORDER BY "CertID"'
            )
            for row in curs:
                vendor_account_id = row['VendorAccountID']
                if vendor_account_id not in result:
                    result[vendor_account_id] = VendorObject(row['VendorName'])

                result[vendor_account_id].add_cert_obj(CertSubObject(
                    row['SubscriptionName'],
                    row['subscriptionID'],
                    row['certName'],
                    row['CertID'],
                    row['CompanyName'],
                    row['AccountID']
                ))

        if len(result) == 0:
            logger.debug("Certificate subscriptions not found")
            return False

        logger.info('\n\nFound the following subscriptions related to the deprecated C++ SSL certificates module:\n')
        for vendor_acc_id, vendor_obj in result.iteritems():
            logger.info(
                'The vendor "%s" (ID #%s) has the following SSL certificate subscriptions in Billing module based on '
                'deprecated C++ SSL certificate plug-ins:\n' % (vendor_obj.name, vendor_acc_id)
            )
            for cert_sub_obj in vendor_obj.subs:
                logger.info(cert_sub_obj.get_cert_info())

        return True

    # Return True if certificate service plans found
    def _find_and_report_cert_plans_info(self):
        if not is_billing_db_table_exists(self.pba_db_connection, 'CertST'):
            logger.debug("Certificate service plans cannot exist: the BSS DB table 'CertST' not found")
            return False

        with self.pba_db_connection.cursor(cursor_factory=RealDictCursor) as curs:
            curs.execute(
                'SELECT p."PlanID", '
                'p."name", '
                'p."AccountID", '
                'a."CompanyName" '
                'FROM "Plan" as p '
                'JOIN "CertST" as cst ON p."serviceTemplateID" = cst."serviceTemplateID" '
                'JOIN "Account" as a ON p."AccountID" = a."AccountID" '
                'ORDER BY p."PlanID"'
            )

            if curs.rowcount == 0:
                logger.debug("Certificate service plans not found")
                return False

            logger.info('\n\nFound the following Service Plans related to the deprecated C++ SSL certificate module:\n')
            for row in curs:
                logger.info(
                    'Service Plan ID %s, Name: "%s" (Vendor ID #%s "%s")' %
                    (str(row['PlanID']), row['name'], str(row['AccountID']), row['CompanyName'])
                )

        return True

    # Return True if certificate service templates found
    def _find_and_report_cert_templates_info(self):
        with self.pba_db_connection.cursor(cursor_factory=RealDictCursor) as curs:
            curs.execute(
                'SELECT st."serviceTemplateID", '
                'st."name", '
                'st."AccountID", '
                'a."CompanyName" '
                'FROM "ServiceTemplate" as st '
                'JOIN "Account" as a ON st."AccountID" = a."AccountID" '
                'WHERE st."GateID" = %d '
                'ORDER BY st."serviceTemplateID"' % CERTGATE_ID
            )

            if curs.rowcount == 0:
                logger.debug("Certificate service templates not found")
                return False

            logger.info(
                '\n\nFound the following Service Templates related to the deprecated C++ SSL certificate module:\n'
            )
            for row in curs:
                logger.info(
                    'Service Template ID %s, Name: "%s" (Vendor ID #%s "%s")' %
                    (str(row['serviceTemplateID']), row['name'], str(row['AccountID']), row['CompanyName'])
                )

        return True

    # Return True if certificate plug-ins found
    def _find_and_report_cert_plugins_info(self):
        if not is_billing_db_table_exists(self.pba_db_connection, 'CertPlugin'):
            logger.debug("Certificate plug-ins cannot exist: the BSS DB table 'CertPlugin' not found")
            return False

        with self.pba_db_connection.cursor(cursor_factory=RealDictCursor) as curs:
            curs.execute(
                'SELECT cp."PluginID", '
                'cp."Name", '
                'cp."VendorID", '
                'a."CompanyName" '
                'FROM "CertPlugin" as cp '
                'JOIN "Account" as a ON cp."VendorID" = a."AccountID" '
                'ORDER BY cp."PluginID"'
            )

            if curs.rowcount == 0:
                logger.debug("Certificate plug-ins not found")
                return False

            logger.info('\n\nFound the following plug-ins related to deprecated C++ SSL certificates module:\n')
            for row in curs:
                logger.info(
                    'Plug-in ID %s, Name: "%s" (Vendor ID #%s "%s")' %
                    (str(row['PluginID']), row['Name'], str(row['VendorID']), row['CompanyName'])
                )

            logger.info('')

        return True

    # Return True if rows found
    def _find_and_report_certgate_rows_bss_db(self, query, db_table_name, check_db_table_existence=False):
        if check_db_table_existence and not is_billing_db_table_exists(self.pba_db_connection, db_table_name):
            logger.debug("Cannot count certgate rows: the BSS DB table '%s' not found" % db_table_name)
            return False

        if 'count' not in query.lower():
            raise ValueError("Unable to execute the DB query '%s': it must contain COUNT operation" % query)

        with self.pba_db_connection.cursor() as curs:
            curs.execute(query)
            if curs.fetchone()[0] > 0:
                logger.info(
                    'The BSS DB table "%s" contains row(s) related to deprecated CERTGATE and/or its plug-ins'
                    % db_table_name
                )
                return True

        return False

    # Return True if files found
    def _find_and_report_cert_files(self):
        is_files_found = False
        for obsolete_file in OBSOLETE_FILES:
            if file_exists(uBilling.PBAConf().get_host_id(), obsolete_file):
                is_files_found = True
                logger.info(
                    "Found the file '%s' related to deprecated CERTGATE and/or its plug-ins on PBAAPPL node"
                    % obsolete_file
                )

        return is_files_found

    def export_certs(self):
        # PRE-CHECK: Necessary DB tables exist
        if not self._is_cert_export_available():
            return

        logger.info("Exporting information about C++ SSL Certificates from Billing DB started")
        filename = DECRYPTED_CERTS_DATA_FILENAME_PREFIX + str(int(time.time())) + ".csv"
        full_file_path = os.path.join(os.path.abspath(os.getcwd()), filename)
        try:
            with self.pba_db_connection.cursor() as curs:
                curs.execute(
                    'SELECT c."CertID", c."subscriptionID", c."Name" as "certName", c."CSR", c."PRIVKEY", ' \
                    's."AccountID", s."SubscriptionName", a."CompanyName", cst."PluginID", ' \
                    'cp."ContainerID", cp."Name" as "certPluginName" ' \
                    'FROM "Cert" as c ' \
                    'JOIN "Subscription" as s ON c."subscriptionID" = s."subscriptionID" ' \
                    'JOIN "Account" as a ON s."AccountID" = a."AccountID" ' \
                    'JOIN "CertST" as cst ON s."serviceTemplateID" = cst."serviceTemplateID" ' \
                    'JOIN "CertPlugin" as cp ON cst."PluginID" = cp."PluginID" ' \
                    'ORDER BY "CertID"'
                )

                with open(filename, "wb") as f:
                    # BOM utf-8
                    f.write(u'\ufeff'.encode('utf8'))
                    logger.info(
                        "\n\nFile '" + full_file_path + "' created.\n\n!!! Important !!!\nMake sure the file is secured"
                        " and not available for non-authorized access because it contains sensitive data.\n"
                    )

                    csv_writer = csv.writer(
                        f, dialect='excel', delimiter=";", lineterminator="\n", quotechar='"', quoting=csv.QUOTE_MINIMAL
                    )
                    # Headers
                    csv_writer.writerow([
                        "Certificate ID", "BSS Subscription ID", "Certificate Name",
                        "Certificate Signing Request (CSR)", "Private Key", "BSS Account ID", "BSS Subscription Name",
                        "Company Name", "Certificate Plugin ID", "Certificate Container ID", "Certificate Plugin Name"
                    ])
                    # Content
                    csv_writer.writerows(curs)

                logger.info("Information about C++ SSL certificates was successfully exported to the file '" + filename)
        except Exception as ex:
            logger.error("Failed to export information about C++ SSL certificates to the file '" + filename)
            raise ex

    def _is_cert_export_available(self):
        check_db_tables = ['Cert', 'CertST', 'CertPlugin']
        for check_db_table in check_db_tables:
            if not is_billing_db_table_exists(self.pba_db_connection, check_db_table):
                logger.info(
                    "Unable to export information about C++ SSL Certificates from Billing DB, the DB table '%s' not "
                    "found." % check_db_table
                )
                return False

        return True

    def clean_certs_subs_orders(self):
        self._user_confirmation(
            "\n\nWARNING!\n\nThis operation can not be undone.\nMake sure you have exported certificates using this "
            "utility with the 'export-cert-data' option.\nAlso, make sure you have created a backup of Billing Database"
            ", namely the following tables if they exist:\n- 'Cert'\n- 'CertArc'\n- 'CertEnomCert'\n- "
            "'CertEnomCertArc'\n- 'OpenSRSCert'\n- 'OpenSRSCertArc'\n- 'GlobalSignCert'\n- 'GlobalSignCertArc'\n- "
            "'SSLStoreCert'\n- 'SSLStoreCertArc'\n\nType 'yes' to continue...\n"
        )

        self._truncate_plugin_specific_db_tables()
        self._clean_certs_subs_orders()

    def archive_docs_plans_templates(self):
        # PRE-CHECK: Check if 'clean-certificates' successfully passed
        with self.pba_db_connection.cursor() as curs_subs:
            curs_subs.execute(
                'SELECT COUNT(*) FROM "Subscription" as s '
                'JOIN "ServiceTemplate" as st ON s."serviceTemplateID" = st."serviceTemplateID" '
                'WHERE st."GateID" = %d' % CERTGATE_ID
            )

            ssl_cert_subscription_counter = curs_subs.fetchone()[0]
            if ssl_cert_subscription_counter > 0:
                raise CleanUpException(
                    "Unable to archive service plans for %s: found %d subscriptions based on these service plans.\n"
                    "You should run this tool using the option 'clean-certificates' as described in the KB article %s."
                    % (BillingModule.CPP_CERT_GATE, ssl_cert_subscription_counter, self.help_link)
                )

        self._user_confirmation(
            "\n\nWARNING!\n\nThis operation can not be undone.\nMake sure you have created a backup of Billing Database"
            ", namely the following tables:\n- 'ARDoc'\n- 'DocDet'\n- 'Plan'\n- 'PlanPeriod'\n\nType 'yes' to "
            "continue...\n"
        )

        # STEP 1: Service Plans and Documents
        success_counter = 0
        fail_counter = 0
        self._init_pba_api()
        with self.pba_db_connection.cursor() as curs_plans:
            curs_plans.execute(
                'SELECT p."PlanID", array_agg(pp."PlanPeriodID") as Periods FROM "Plan" as p '
                'JOIN "ServiceTemplate" as st ON p."serviceTemplateID" = st."serviceTemplateID" '
                'JOIN "PlanPeriod" pp ON p."PlanID" = pp."PlanID" '
                'WHERE st."GateID" = %d '
                'GROUP BY p."PlanID" ORDER BY p."PlanID" DESC' % CERTGATE_ID
            )

            for row in curs_plans:
                service_plan_id = row[0]
                plan_period_ids = row[1]
                try:
                    self._archive_plan_with_documents(service_plan_id, plan_period_ids)
                    success_counter += 1
                except Exception as ex:
                    fail_counter += 1
                    logger.error(
                        "\tFailed to archive the service plan with ID %d or its documents. The error: %s\n"
                        % (service_plan_id, ex)
                    )
                    logger.error(traceback.format_exc())

        logger.info(
            "\n\nCleaning for service plans finished. Successfully archived %d of %d service plans.\n"
            % (success_counter, success_counter + fail_counter)
        )

        if fail_counter > 0:
            raise CleanUpException(
                "Failed to archive %d service plans. Fix the problem(s) and re-run the clean-up tool." % fail_counter
            )

        # STEP 2: Service Templates
        success_counter = 0
        fail_counter = 0
        with self.pba_db_connection.cursor() as curs_templates:
            curs_templates.execute(
                'SELECT st."serviceTemplateID" FROM "ServiceTemplate" as st '
                'WHERE st."GateID" = %d '
                'ORDER BY st."serviceTemplateID" DESC' % CERTGATE_ID
            )

            for row_template in curs_templates:
                service_template_id = row_template[0]
                try:
                    self._archive_service_template(service_template_id)
                    success_counter += 1
                except Exception as ex:
                    fail_counter += 1
                    logger.error(
                        "Failed to archive the billing service template with ID %d. The error: %s\n"
                        % (service_template_id, ex)
                    )
                    logger.error(traceback.format_exc())

        logger.info(
            "\n\nCleaning for service templates finished. Successfully archived %d of %d service templates.\n"
            % (success_counter, success_counter + fail_counter)
        )

        if fail_counter > 0:
            raise CleanUpException(
                "Failed to archive %d service templates. Fix the problem(s) and re-run the clean-up tool."
                % fail_counter
            )

    def _archive_service_template(self, st_id):
        logger.info("Archiving the billing service template with ID '%s' started." % str(st_id))
        logger.debug(
            "Calling BSS API method 'BM.ServiceTemplateMoveToArc' for the billing service template ID '%s'" % str(st_id)
        )
        if not DRY_RUN:
            result = self.pba_api.BM.ServiceTemplateMoveToArc(st_id)
            logger.debug("Result: " + str(result))

        logger.info("The billing service template with ID '%s' successfully moved to archive.\n" % str(st_id))

    def _archive_plan_with_documents(self, plan_id, plan_period_ids):
        logger.info(
            "Archiving the service plan with ID '%s' and its documents related to the service plan's periods with IDs "
            "%s." % (str(plan_id), ', '.join(map(str, plan_period_ids)))
        )

        self._archive_documents(plan_period_ids)

        # Unblock delegated service plans for archiving
        self._delete_query_bss_db(
            'DELETE FROM "ResellerTrans" WHERE "PlanPeriodID" IN (\'%s\')' % "','".join(map(str, plan_period_ids))
        )

        # Archive orders indirectly related to SSL plans, e.g. orders for reseller's and certificate's plans together
        self._archive_plan_periods_orders(plan_period_ids)

        logger.debug("\tCalling BSS API method 'BM.PlanMoveToArc' for the service plan ID '%s'" % str(plan_id))
        if not DRY_RUN:
            result = self.pba_api.BM.PlanMoveToArc(plan_id)
            logger.debug("\tResult: " + str(result))

        logger.info(
            "The service plan with ID '%s' and its documents successfully moved to archive.\n" % str(plan_id)
        )

    def _archive_documents(self, plan_period_ids):
        with self.pba_db_connection.cursor() as curs:
            curs.execute(
                'SELECT DISTINCT dd."DocID" FROM "DocDet" as dd '
                'WHERE dd."PlanPeriodID" IN (%s) ORDER BY dd."DocID" DESC' % ','.join(map(str, plan_period_ids))
            )

            for doc_detail_row in curs:
                doc_id = doc_detail_row[0]
                logger.info("\tArchiving the billing document with ID '%s' started." % str(doc_id))
                logger.debug(
                    "\tCalling BSS API method 'BM.ARDocMoveToArc' for the billing document ID '%s'" % str(doc_id)
                )
                if not DRY_RUN:
                    result = self.pba_api.BM.ARDocMoveToArc(doc_id)
                    logger.debug("\tResult: " + str(result))

                logger.info("\tThe billing document with ID '%s' successfully moved to archive.\n" % str(doc_id))

    def clean_certgate(self):
        # PRE-CHECK: Check if 'clean-service-plans' successfully passed and cert plug-ins were deleted in BSS PCP/RCP
        self._clean_certgate_pre_check()

        affected_db_tables = [
            'AEvent', 'AEventKind', 'CertPluginTemplate', 'CertProduct', 'CertSetup', 'DBContainers', 'Method',
            'RegisteredServer', 'Scheduler', 'ServiceGate', 'Task', 'TaskHandler', 'VendorDefaultConfig'
        ]
        self._user_confirmation(
            "\n\nWARNING!\n\nThis operation can not be undone.\nMake sure you have created a backup of Billing Database"
            ", namely the following tables:%s\n\nType 'yes' to continue...\n"
            % ''.join("\n- '%s'" % el for el in affected_db_tables)
        )

        # Clean DB tables to pass the BM pre-check '21-00-eol-ssl-certificates-pre-check.py'
        if is_billing_db_table_exists(self.pba_db_connection, 'CertPluginTemplate'):
            # Recursively delete dependant rows in DB table "CertProduct"
            self._delete_query_bss_db('DELETE FROM "CertPluginTemplate"')

        self._delete_query_bss_db(
            'DELETE FROM "DBContainers" WHERE "Name" IN (\'%s\')' % \
            "','".join(map(lambda p: p + " " * (80 - len(p)), DEPRECATED_CERT_CONTAINERS))
        )

        self._delete_query_bss_db(
            'DELETE FROM "RegisteredServer" WHERE "SrvID" = %d AND "ContName" = \'CERTGATE\'' % CERTGATE_ID
        )

        self._delete_query_bss_db(
            'DELETE FROM "ServiceGate" WHERE "GateID" = %d AND "ContName" = \'CERTGATE\'' % CERTGATE_ID
        )

        self._delete_query_bss_db(
            'DELETE FROM "Method" WHERE "Container" IN (\'%s\')' % "','".join(DEPRECATED_CERT_CONTAINERS)
        )

        self._delete_query_bss_db('DELETE FROM "VendorDefaultConfig" WHERE "Container" = \'CERTGATE\'')
        self._delete_query_bss_db('DELETE FROM "CertSetup"')

        self._delete_billing_events()

        logger.debug("Deleting the files:%s" % ''.join("\n- %s" % el for el in OBSOLETE_FILES))
        if not DRY_RUN:
            remove_files(uBilling.PBAConf().get_host_id(), OBSOLETE_FILES, True)

        logger.info("\n\nCleaning CERTGATE for BSS DB successfully finished.\n")

    def _delete_billing_events(self):
        # 1. Scheduler - stop generating new periodic events and tasks
        self._delete_query_bss_db('DELETE FROM "Scheduler" WHERE "EKID" = \'Periodic check for certificates statuses\'')

        # 2. Delete CERTGATE TaskHandlers - their implementations are already deleted at this moment
        self._delete_query_bss_db(
            'DELETE FROM "TaskHandler" WHERE "EKID" = \'Erase Account Data\' AND "ObjName" = \'CERTGATE\' AND '
            '"MethodName" = \'HandleAccountErase\''
        )

        self._delete_query_bss_db(
            'DELETE FROM "TaskHandler" WHERE "EKID" = \'Provider Initialized\' AND "ObjName" = \'CERTGATE\' AND '
            '"MethodName" = \'SetDefaultConfiguration\''
        )

        self._delete_query_bss_db(
            'DELETE FROM "TaskHandler" WHERE "EKID" = \'Periodic check for certificates statuses\' AND "ObjName" = '
            '\'CERTGATE\' AND "MethodName" = \'CheckCertStatusesTask\''
        )

        # 3. Delete failed tasks (status=5) to unblock deletion for the related failed AEvents (status=2)
        self._delete_query_bss_db(
            'DELETE FROM "Task" as t USING "AEvent" as ae WHERE t."EID" = ae."EID" AND ae."EKID" = '
            '\'Periodic check for certificates statuses\' AND t."TaskStat" = 5 AND ae."Status" = 2'
        )

        # 4. Delete AEvents
        self._delete_query_bss_db('DELETE FROM "AEvent" WHERE "EKID" = \'Periodic check for certificates statuses\'')

        # 5. Delete AEventType
        self._delete_query_bss_db(
            'DELETE FROM "AEventKind" WHERE "EKID" = \'Periodic check for certificates statuses\''
        )

    def _delete_query_bss_db(self, cmd):
        if 'delete' not in cmd.lower():
            raise ValueError("Unable to execute the DB query '%s': it must contain DELETE operation" % cmd)

        with self.pba_db_connection.cursor() as curs:
            logger.debug(cmd)
            if not DRY_RUN:
                curs.execute(cmd)
                self.pba_db_connection.commit()

    def _clean_certgate_pre_check(self):
        if is_billing_db_table_exists(self.pba_db_connection, 'CertPlugin'):
            with self.pba_db_connection.cursor() as curs_plugins:
                curs_plugins.execute('SELECT COUNT(*) FROM "CertPlugin"')
                ssl_cert_plugin_counter = curs_plugins.fetchone()[0]
                if ssl_cert_plugin_counter > 0:
                    raise CleanUpException(
                        "Unable to clean CERTGATE for %s: found %d instance(s) of C++ SSL certificate plug-ins.\nYou "
                        "should run this tool using the option 'clean-service-plans' and delete all instances of C++ "
                        "SSL certificate plug-ins using PCP/RCP as described in the KB article %s."
                        % (BillingModule.CPP_CERT_GATE, ssl_cert_plugin_counter, self.help_link)
                    )

    def _clean_certs_subs_orders(self):
        success_counter = 0
        fail_counter = 0
        self._init_pba_api()
        with self.pba_db_connection.cursor() as curs_subs:
            curs_subs.execute(
                'SELECT s."subscriptionID" FROM "Subscription" as s '
                'JOIN "ServiceTemplate" as st ON s."serviceTemplateID" = st."serviceTemplateID" '
                'WHERE st."GateID" = %d ORDER BY s."subscriptionID" DESC' % CERTGATE_ID
            )

            for row_sub in curs_subs:
                cert_sub_id = row_sub[0]
                try:
                    self._archive_sub_with_orders(cert_sub_id)
                    success_counter += 1
                except Exception as ex:
                    fail_counter += 1
                    logger.error(
                        "\tFailed to archive the billing subscription with ID %d or its orders. The error: %s\n"
                        % (cert_sub_id, ex)
                    )
                    logger.error(traceback.format_exc())

        logger.info(
            "\n\nCleaning for C++ SSL certificates finished. Successfully archived %d of %d subscriptions.\n"
            % (success_counter, success_counter + fail_counter)
        )

        if fail_counter > 0:
            raise CleanUpException(
                "Failed to archive %d subscriptions. Fix the problem(s) and re-run the clean-up tool." % fail_counter
            )

    def _archive_sub_with_orders(self, sub_id):
        logger.info("Archiving the billing subscription with ID '%s' and its orders started." % str(sub_id))

        self._archive_sub_orders(sub_id)

        logger.debug(
            "\tCalling BSS API method 'BM.SubscriptionMoveToArc' for the billing subscription ID '%s'" % str(sub_id)
        )
        if not DRY_RUN:
            result = self.pba_api.BM.SubscriptionMoveToArc(sub_id)
            logger.debug("\tResult: " + str(result))

        logger.info(
            "The billing subscription with ID '%s' and its orders successfully moved to archive.\n" % str(sub_id)
        )

    def _archive_sub_orders(self, sub_id):
        with self.pba_db_connection.cursor() as curs_oitems:
            curs_oitems.execute(
                'SELECT o."OrderDocOrderID" FROM "OItem" as o '
                'WHERE o."subscriptionID" = %d ORDER BY o."OrderDocOrderID" DESC' % sub_id
            )

            for row_oitem in curs_oitems:
                order_id = row_oitem[0]
                self._archive_order(order_id)

    def _archive_order(self, order_id):
        logger.info("\tArchiving the billing order with ID '%s' started." % str(order_id))
        logger.debug("\tCalling BSS API method 'BM.SalesOrderMoveToArc' for the billing order ID '%s'" % str(order_id))
        if not DRY_RUN:
            result = self.pba_api.BM.SalesOrderMoveToArc(order_id)
            logger.debug("\tResult: " + str(result))

        logger.info("\tThe billing order with ID '%s' successfully moved to archive.\n" % str(order_id))

    def _archive_plan_periods_orders(self, plan_period_ids):
        with self.pba_db_connection.cursor() as curs_order_details:
            curs_order_details.execute(
                'SELECT DISTINCT od."OrderID", so."OrderStatusID" FROM "OrdDet" as od '
                'JOIN "SalesOrder" as so ON od."OrderID" = so."OrderID" '
                'WHERE od."PlanPeriodID" IN (\'%s\') ORDER BY od."OrderID" DESC' % "','".join(map(str, plan_period_ids))
            )

            for row_order_detail in curs_order_details:
                order_id = row_order_detail[0]
                order_status_id = row_order_detail[1]
                try:
                    self._archive_order(order_id)
                except Fault as fault:
                    if 'The Billing Order cannot be deleted before it is invoiced. Please try to delete the order ' \
                       'after the invoice is created.' not in fault.faultString or order_status_id != 'CP':
                        raise fault

                    # Try a workaround for 'Completed' orders without invoice (https://jira.int.zone/browse/OA-3718)
                    # Ignore 'Total_Value' because it can be non-zero and zero (https://jira.int.zone/browse/OA-27243)
                    logger.warning(
                        'Unable to archive the order #%d because it does not have an invoice. Applying the workaround: '
                        'mark the order as technical and re-try archiving.' % order_id
                    )
                    with self.pba_db_connection.cursor() as cur:
                        cmd = 'UPDATE "SalesOrder" SET "IsTechOrder" = 1 WHERE "OrderID" = %d' % order_id
                        logger.debug(cmd)
                        cur.execute(cmd)

                        cmd = 'UPDATE "ResellerTrans" as rt SET "ReselOrdDetID" = null FROM "OrdDet" as od ' \
                              'WHERE od."OrderID" = %d AND od."DetID" = rt."ReselOrdDetID"' % order_id
                        logger.debug(cmd)
                        cur.execute(cmd)

                        self.pba_db_connection.commit()

                    # Retry archiving for the failed order after applying the workaround
                    self._archive_order(order_id)
                    logger.info(
                        'Workaround was successfully applied and order #%d was successfully archived.' % order_id
                    )

    def _truncate_plugin_specific_db_tables(self):
        logger.info("Removing certificates from plugin-specific BSS DB tables started.")
        found_tables_for_truncating = False
        with self.pba_db_connection.cursor() as cur:
            for cert_plugin_table in DEPRECATED_DB_TABLES_REGISTRAR_CERT_INFO:
                if is_billing_db_table_exists(self.pba_db_connection, cert_plugin_table):
                    found_tables_for_truncating = True
                    cmd = 'TRUNCATE TABLE "%s"' % cert_plugin_table
                    logger.debug(cmd)
                    if not DRY_RUN:
                        cur.execute(cmd)

            if found_tables_for_truncating and not DRY_RUN:
                self.pba_db_connection.commit()

        logger.info("Plugin-specific BSS DB tables were successfully cleaned.\n")

    def _init_pba_api(self):
        pbaapi.initFromEnv()
        self.pba_api = pbaapi.PBAAPIRaw()

    def _user_confirmation(self, message):
        logger.warning(message)
        confirm = raw_input().lower()
        if confirm.lower() != 'yes':
            raise CleanUpException("Operation aborted by user")
