# (C)2021 Ingram Micro Inc. all Rights Reserved
from poaupdater import uBilling

from oacleanup.cleanup.base import CleanUpProcedureBase
from oacleanup.settings import Step, formatter, BillingModule
from oacleanup.utils import find_absent_values, find_invalide_values, find_duplicates
from oacleanup.utils.billing import is_billing_installed
from oacleanup.utils.billing_db import get_connection_to_billing_db
from oacleanup.utils.hosts import remove_files
from oacleanup.utils.errors import CleanUpException

import logging
import encodings.idna

logger = logging.getLogger()

debug_mode = False
help_link = 'https://cloudblue.freshdesk.com/a/solutions/articles/44002194907'
deprecated_plugins = [
    'ENOM', 'GODADDY', 'BOTTLE', 'DKHOSTMASTER', 'DNSBE', 'EPNIC', 'EPPREGISTRAR', 'EURID', 'INTERNETX', 'IRELANDDR',
    'JOKER', 'MIT', 'NOMINET', 'OPENSRS', 'DOMINIC', 'RUCENTER', 'SIDN', 'UNIFORUM', 'WHOIS', 'KAZNIC', 'WEBNIC',
    'DISTRIBUTEIT', 'AUSREGISTRY'
]
allowed_plugins = ['PEM_REGISTRAR', 'DUMMY_REGISTRAR', 'TEST_REGISTRAR']

# Some plugin's DB tables look generic and will not be dropped for safety reasons
deprecated_bss_db_tables = [
    # OPENSRS
    'OpenSRSAutoRenewTLD', 'OpenSRSConf', 'OpenSRSConfArc', # 'DBPatches',
    # AUSREGISTRY
    'AusRegistryConf', 'AusRegistryContact', 'AusRegTransferConfirm',
    # BOTTLE
    'BottleConf', 'BottleConfArc', 'BottleMember', 'BottleMemberArc',
    # DISTRIBUTEIT
    'DistributeITConf', 'DistributeITConfArc',
    # DKHOSTMASTER
    'DKHostmasterConf', 'DKHostmasterConfArc', 'DKHostmasterNS', 'DKHostmasterNSArc',
    # DNSBE
    'DNSbeConf', 'DNSbeConfArc',
    # DOMINIC
    'DomiNICConf', 'DomiNICCountry', 'DomiNICDomain', # 'ParentCustomer',
    # ENOM
    'EnomConf', 'EnomStateJP',
    # EPNIC
    'EpnicConf', 'EpnicConfArc', 'AnEpnicContact', 'AnEpnicContactArc', 'AnEpnicDomain', 'AnEpnicDomainArc',
    'AnEpnicDomainTaskQueue', 'AnEpnicDomainTaskQueueArc', 'AnEpnicDomainTransferRequest',
    'AnEpnicDomainTransferRequestArc', 'AnEpnicDomainZoneProperty',
    # EPPREGISTRAR
    'EPPRegistrarConf', 'EPPRegistrarConfArc', 'EPPRegistrarContact',
    # EURID
    'EURidConf', 'EURidConfArc',
    # GODADDY
    'GoDaddyConf', 'GoDaddyCountry',
    # INTERNETX
    'InterNetXConf', 'InterNetXContact', 'InterNetXTransaction',
    # IRELANDDR
    'IrelandDRConf', 'IrelandDRConfArc', 'IrelandDRCountry', 'IrelandDRCountryArc',
    # JOKER
    'JokerConf', 'JokerConfArc', 'JokerUser', 'JokerUserArc', 'AnJokerDomain', 'AnJokerDomainArc',
    # KAZNIC
    'KAZNICConf',
    # MIT
    'MelbourneITConf', 'MelbourneITConfArc', 'MITIsTransferInternal',
    # NOMINET
    'NominetAccount', 'NominetConf',
    # RUCENTER
    'RuCenterConf', 'RuCenterContract', 'RuCenterExpDate', 'RuCenterExpDateArc',
    # SIDN
    'SIDNConf', 'SIDNConfArc', 'SIDNContact',
    # UNIFORUM
    'UniForumConf', 'UniForumConfArc',
    # WEBNIC
    'WebNICConf', 'WebNICStateBook',
    # WHOIS
    'WhoisConf'
]

obsolete_customization_symlinks = [
    '/usr/local/bm/conf/wnd/BM/customization.EPNIC'
    '/usr/local/bm/conf/wnd/DOMAINGATE/customization.DOMINIC'
    '/usr/local/bm/conf/wnd/DOMAINGATE/customization.WHOIS'
]

SHOW_PROBLEMS_STEP = Step(
    name='show-problems',
    description=formatter('Show all objects related to the deprecated {module} which must be migrated before cleaning up'),
    function='check_deprecated_registrars_not_used',
    checks=[],
    arguments={}
)

CLEAN_ACCOUNT_DOMAINS_STEP = Step(
    name='account-domains',
    description=formatter('Delete all "Canceled" Domains from Billing module for certain accounts'),
    function='clean_account_domains',
    checks=[],
    arguments={
        '--account-ids': 'Account IDs in Billing for which the "Canceled" Domains must be deleted. '
            'Examples: "--account-ids 1", "--account-ids 10,11,12"',
        '--recursive': 'Possible values: True or False. If True, the canceled domains will be deleted for the specified '
            'accounts and their sub-accounts. Examples: "--recursive True", "--recursive False"'
    }
)

CLEAN_REGISTRAR_DOMAINS_STEP = Step(
    name='registrar-domains',
    description=formatter('Delete all "Canceled" Domains from Billing module for certain Domain registrar instances'),
    function='clean_registrar_domains',
    checks=[],
    arguments={
        '--registrar-ids': 'Domain Registrar IDs in Billing for which the "Canceled" Domains must be deleted. '
            'Examples: "--registrar-ids 5", "--registrar-ids 10,11"'
    }
)

CLEAN_PLUGINS_STEP = Step(
    name='plugins',
    description=formatter('Delete containers of the deprecated C++ domain plug-ins: %s' % (', '.join(deprecated_plugins))),
    function='clean_plugins',
    checks=[],
    arguments={}
)


class CleanUpCppDomainRegistrars(CleanUpProcedureBase):
    def __init__(self):
        super(CleanUpCppDomainRegistrars, self).__init__(
            name=BillingModule.CPP_DOMAIN_REGISTRARS,
            list_of_available_steps=[
                SHOW_PROBLEMS_STEP, CLEAN_ACCOUNT_DOMAINS_STEP, CLEAN_REGISTRAR_DOMAINS_STEP, CLEAN_PLUGINS_STEP
            ],
        )
        self.help_link = help_link
        self.billing_is_installed = is_billing_installed()
        if self.billing_is_installed:
            self.pba_db_connection = get_connection_to_billing_db()

    @staticmethod
    def get_idn_domain_name(name):
        return name if name.find("xn--") < 0 else encodings.idna.ToUnicode(name)

    def check_cpp_plugins_used_in_zones(self):
        with self.pba_db_connection.cursor() as curs:
            curs.execute(
                'SELECT z."DomainZoneID", z."Vendor_AccountID", a."CompanyName" FROM "DomainZone" as z '
                'JOIN "Registrar" as r ON (z."RegistrarID" = r."RegistrarID") '
                'JOIN "Account" as a ON (z."Vendor_AccountID" = a."AccountID") '
                'WHERE r."ContainerID" NOT IN '
                '(\'%s\')' % "','".join(allowed_plugins)
            )
            zones_result = curs.fetchall()
        deprecated_zones_count = len(zones_result)
        zones_map = dict()  # key: vendor ID, value: list of zones
        accounts_map = dict()
        for (domain_zone_id, vendor_account_id, vendor_name) in zones_result:
            accounts_map[vendor_account_id] = vendor_name
            if vendor_account_id not in zones_map.keys():
                zones_map[vendor_account_id] = list()
            zones_map[vendor_account_id].append(domain_zone_id)
        if deprecated_zones_count > 0:
            logger.info("\n\nInfo about zones:\n")
        for vendor_id in zones_map.keys():
            logger.info("The vendor #%s (\"%s\") has the following zones using registrars "
                        "based on deprecated C++ domain plugins: %s" %
                        (vendor_id, accounts_map[vendor_id].decode('utf-8'), ", ".join(
                            map(self.get_idn_domain_name, zones_map[vendor_id]))))
        return deprecated_zones_count

    @staticmethod
    def __fill_domain_related_maps(accounts_map, domains_map, domains_result, registrars_map):
        for (
                domain_id, full_domain_name, tech_domain_name,
                registrar_name, registrar_id,
                vendor_account_id, vendor_name
        ) in domains_result:
            accounts_map[vendor_account_id] = vendor_name
            registrars_map[registrar_name] = registrar_id
            if vendor_account_id not in domains_map.keys():
                domains_map[vendor_account_id] = dict()
            vendor_map = domains_map[vendor_account_id]
            if registrar_name not in vendor_map.keys():
                vendor_map[registrar_name] = list()
            vendor_map[registrar_name].append(
                dict(id=domain_id, name=full_domain_name, tech_domain_name=tech_domain_name))

    def clean_plugins(self):
        logger.info("Cleaning deprecated C++ domain plug-ins started")
        # Check 1
        with self.pba_db_connection.cursor() as cur:
            cur.execute('SELECT COUNT(*) FROM "Registrar" WHERE "ContainerID" IN (\'%s\')' % "','".join(deprecated_plugins))
            active_legacy_instances_number = cur.fetchone()[0]
            if int(active_legacy_instances_number) > 0:
                raise CleanUpException(
                    "There are %d active instances of deprecated Domain plug-ins. You must migrate all domains "
                    "associated with deprecated Domain plug-ins to PHP domain plug-ins first, then delete active "
                    "instances of deprecated Domain plug-ins and re-run the command. For more information, refer to %s."
                    % (int(active_legacy_instances_number), help_link)
                )

        # User confirmation
        self._user_confirmation(
            "\n\nWARNING!\n\nThis operation will drop the following tables in Billing database: '%s'.\n\nDeprecated "
            "records will be deleted in the following Billing DB: 'Plugin', 'DBContainers'.\n\nThis operation can not "
            "be undone. Make sure you have created backups for existing Billing DB table from the mentioned list. "
            "Type 'yes' to continue...\n" % "', '".join(deprecated_bss_db_tables)
        )

        # Cleaning step 1
        logger.info("Step 1: Cleaning up the 'Plugin' Billing DB table")
        with self.pba_db_connection.cursor() as cur:
            cmd = 'DELETE FROM "Plugin" WHERE "ContainerID" IN (\'%s\')' % "','".join(deprecated_plugins)
            logger.info("[DEBUG] %s" % cmd)
            if not debug_mode:
                cur.execute(cmd)
                self.pba_db_connection.commit()

        # Cleaning step 2
        logger.info("Step 2: Cleaning up the 'DBContainers' Billing DB table")
        with self.pba_db_connection.cursor() as cur:
            cmd = 'DELETE FROM "DBContainers" WHERE "Name" IN (\'%s\')' % \
                  "','".join(map(lambda p: p + " " * (80 - len(p)), deprecated_plugins))
            logger.info("[DEBUG] %s" % cmd)
            if not debug_mode:
                cur.execute(cmd)
                self.pba_db_connection.commit()

        # Cleaning step 3
        logger.info("Step 3: Drop the deprecated tables in Billing DB: %s" % ', '.join(deprecated_bss_db_tables))
        pba_verify_con = get_connection_to_billing_db('pba_verify')
        for table in deprecated_bss_db_tables:
            with self.pba_db_connection.cursor() as cur:
                cmd = 'DROP TABLE IF EXISTS "%s"' % table
                logger.info("[DEBUG][db=pba] %s" % cmd)
                if not debug_mode:
                    cur.execute(cmd)
                    self.pba_db_connection.commit()
            with pba_verify_con.cursor() as cur_vrf:
                cmd = 'DROP TABLE IF EXISTS "%s"' % table
                logger.info("[DEBUG][db=pba_verify] %s" % cmd)
                if not debug_mode:
                    cur_vrf.execute(cmd)
                    pba_verify_con.commit()
        remove_files(uBilling.PBAConf().get_host_id(), obsolete_customization_symlinks)

        logger.info("Deprecated C++ domain plugins were successfully deleted")

    def clean_registrar_domains(self, registrar_ids):
        logger.info("Started verifying for passed parameters")
        registrar_ids = registrar_ids.split(',')
        # Check 1
        duplicates = list(find_duplicates(registrar_ids))
        if len(duplicates) > 0:
            raise CleanUpException(
                "Duplicate registrar IDs were specified: %s. Specify unique identifiers."
                % ", ".join([str(item) for item in duplicates])
            )

        # Check 2
        invalid_values = list(find_invalide_values(registrar_ids))
        if len(invalid_values) > 0:
            raise CleanUpException(
                "Invalid registrar IDs were specified: %s. Specify the IDs of existing registrar instances."
                % ", ".join([str(item) for item in invalid_values])
            )

        registrar_ids = [int(x) for x in registrar_ids]
        # Check 3
        absent_registrars = find_absent_values(
            registrar_ids, 'SELECT "RegistrarID" FROM "Registrar" WHERE "RegistrarID" IN %s', True
        )
        if len(absent_registrars) > 0:
            raise CleanUpException(
                "There are no C++ domain registrar instances with the following IDs: %s. Specify existing IDs."
                % ", ".join([str(item) for item in absent_registrars])
            )

        # Check 4
        for registrar_id in registrar_ids:
            with self.pba_db_connection.cursor() as cur:
                cur.execute('SELECT "ContainerID" FROM "Registrar" WHERE "RegistrarID" = %d' % registrar_id)
                container_id = cur.fetchone()[0]
                if container_id not in deprecated_plugins:
                    raise CleanUpException(
                        "Invalid Registrar's instance ID #%d (%s) specified. Specify only deprecated Registrar's instance IDs."
                        % (registrar_id, container_id)
                    )

        logger.info("Passed parameters are valid.")

        # User confirmation
        self._user_confirmation(
            "\n\nWARNING!\n\nThis operation can not be undone. Make sure you have created a backup for the 'ADomain' "
            "Billing DB table. Type 'yes' to continue...\n"
        )

        # Cleaning
        logger.info("Cleaning 'Canceled' Billing Domains by Registrar IDs started")
        for registrar_id in registrar_ids:
            self._delete_registrar_canceled_domain(registrar_id)

        logger.info("'Canceled' Domains were successfully deleted from Billing DB")

    def clean_account_domains(self, account_ids, recursive):
        logger.info("Started verifying for passed parameters")
        account_ids = account_ids.split(',')
        # Check 1
        duplicates = list(find_duplicates(account_ids))
        if len(duplicates) > 0:
            raise CleanUpException(
                "Duplicate account IDs were specified: %s. Specify unique identifiers."
                % ", ".join([str(item) for item in duplicates])
            )

        # Check 2
        invalid_values = list(find_invalide_values(account_ids))
        if len(invalid_values) > 0:
            raise CleanUpException(
                "Invalid account IDs were specified: %s. Specify the IDs of existing accounts."
                % ", ".join([str(item) for item in invalid_values])
            )

        account_ids = [int(x) for x in account_ids]
        # Check 3
        absent_accounts = find_absent_values(
            account_ids, 'SELECT "AccountID" FROM "Account" WHERE "AccountID" IN %s', True
        )
        if len(absent_accounts) > 0:
            raise CleanUpException(
                "There are no accounts in Billing with the following IDs: %s. Specify the IDs of existing accounts."
                % ", ".join([str(item) for item in absent_accounts])
            )

        # Check 4
        if recursive.lower() not in ['true', 'false']:
            raise CleanUpException("Invalid value for the '--recursive' option. Allowed values: True or False.")

        recursive = recursive.lower() == 'true'
        logger.info("Passed parameters are valid.")

        # User confirmation
        self._user_confirmation(
            "\n\nWARNING!\n\nThis operation can not be undone. Make sure you have created a backup for the 'ADomain' "
            "Billing DB table. Type 'yes' to continue...\n"
        )

        # Delete Canceled Domains
        logger.info("Cleaning 'Canceled' Billing Domains by Account IDs started")
        for parent_account_id in account_ids:
            if recursive:
                for children_account_id in self._get_accounts(parent_account_id):
                    self._delete_account_canceled_domain(children_account_id)
            else:
                self._delete_account_canceled_domain(parent_account_id)

        logger.info("'Canceled' Domains were successfully deleted from Billing DB")

    def check_deprecated_registrars_not_used(self):
        if not self.billing_is_installed:
            logger.info("Billing is not installed, skip the check")
            return
        deprecated_zones_count = self.check_cpp_plugins_used_in_zones()
        with self.pba_db_connection.cursor() as curs:
            curs.execute('SELECT d."DomainID", d."FullDomainName", d."TechDomainName", r."Name" as registrarname, '
                         'r."RegistrarID" as registrarid, r."Vendor_AccountID", a."CompanyName" FROM "ADomain" as d '
                         'JOIN "Registrar" as r ON (d."RegistrarID" = r."RegistrarID") '
                         'JOIN "Account" as a ON (r."Vendor_AccountID" = a."AccountID")'
                         'WHERE r."ContainerID" NOT '
                         'IN (\'%s\')' % "','".join(allowed_plugins))
            domains_result = curs.fetchall()

        deprecated_domains_count = len(domains_result)
        domains_map = dict()
        registrars_map = dict()
        accounts_map = dict()
        self.__fill_domain_related_maps(accounts_map, domains_map, domains_result, registrars_map)

        if deprecated_domains_count > 0:
            logger.info("\n\nInfo about registered domains:\n")
        for vendor_id in domains_map.keys():
            logger.info(
                "The vendor #%s (%s) has the following domains registered by registrars "
                "based on deprecated C++ domain plugins:" % (vendor_id, accounts_map[vendor_id].decode('utf-8')))
            for registrar_name in domains_map[vendor_id].keys():
                logger.info("   On registrar \"%s\" (ID=%s):", registrar_name, registrars_map[registrar_name])
                for domain in domains_map[vendor_id][registrar_name]:
                    full_domain_name = domain["name"] if domain["name"] == domain["tech_domain_name"] \
                        else "%s (%s)" % (domain["name"], domain["tech_domain_name"])
                    logger.info("        Billing Domain ID: %s, Name: %s" % (domain["id"], full_domain_name))

        # Show unused deprecated Registrars (without domains and associated TLDs) including WHOIS plug-in
        with self.pba_db_connection.cursor() as curs:
            curs.execute('SELECT r."RegistrarID", r."Name", r."ContainerID", r."Vendor_AccountID", a."CompanyName" '
                         'FROM "Registrar" as r JOIN "Account" as a ON (r."Vendor_AccountID" = a."AccountID") '
                         'WHERE "ContainerID" NOT IN (\'%s\')' % "','".join(allowed_plugins))
            deprecated_registrar_instances = curs.fetchall()
        deprecated_registrars_count = len(deprecated_registrar_instances)

        if deprecated_registrars_count > 0:
            logger.info("\n\nInfo about deprecated registrars:\n")
        for registrar_id, registrar_name, container_id, acc_id, company_name in deprecated_registrar_instances:
            logger.info(
                "The vendor #%s (%s) has the active instance '%s' (ID #%s) of the deprecated registrar '%s'",
                acc_id, company_name, registrar_name, registrar_id, container_id
            )

        if bool(deprecated_zones_count != 0 or deprecated_domains_count != 0 or deprecated_registrars_count != 0):
            logger.info("To migrate from deprecated C++ domain registrar plugins please refer to KB: %s" %
                        self.help_link)
        else:
            logger.info("Deprecated C++ domain registrar plugins are not used in the system.")

    def _get_accounts(self, parent_account_id):
        result = [parent_account_id]
        with self.pba_db_connection.cursor() as cur:
            cur.execute('SELECT "AccountID" FROM "Account" WHERE "VendorAccountID" = %d' % parent_account_id)
            for row in cur:
                result.extend(self._get_accounts(row[0]))

        return result

    def _delete_registrar_canceled_domain(self, registrar_id):
        logger.info("Deleting 'Canceled' Billing Domains for the Registrar ID %d" % registrar_id)
        with self.pba_db_connection.cursor() as cur:
            cmd = 'DELETE FROM "ADomain" WHERE "Status" = 380 AND "RegistrarID" = %d' % registrar_id
            logger.info("[DEBUG] %s" % cmd)
            if not debug_mode:
                cur.execute(cmd)
                self.pba_db_connection.commit()

        logger.info("'Canceled' Billing Domains successfully deleted for the Registrar ID %d" % registrar_id)

    def _delete_account_canceled_domain(self, account_id):
        logger.info("Deleting 'Canceled' Billing Domains for the Account ID %d" % account_id)
        with self.pba_db_connection.cursor() as cur:
            cmd = 'DELETE FROM "ADomain" as d USING "Registrar" as r WHERE d."RegistrarID" = r."RegistrarID" AND ' \
                  'd."Status" = 380 AND d."RegisteredTo_AccountID" = %d AND r."ContainerID" IN (\'%s\') ' \
                  % (account_id, "','".join(deprecated_plugins))
            logger.info("[DEBUG] %s" % cmd)
            if not debug_mode:
                cur.execute(cmd)
                self.pba_db_connection.commit()

        logger.info("'Canceled' Billing Domains successfully deleted for the Account ID %d" % account_id)

    def _user_confirmation(self, message):
        logger.info(message)
        confirm = raw_input().lower()
        if confirm.lower() != 'yes':
            raise CleanUpException("Operation aborted by user")
