# (C)2018 Ingram Micro Inc. all Rights Reserved
from sh_base import CleanUpSHBase

from oacleanup.settings import OAModule, CleanUpSteps
from oacleanup.utils import packages, hosts
from oacleanup.utils.billing import is_billing_installed, get_billing_api

from poaupdater import uSysDB, openapi, uPEM, uTasks

import logging

logger = logging.getLogger()

EXCHANGE_SERVICES = [
    ('service', 'Exchange2010OAB'),
    ('service', 'Exchange2010IMAP4'),
    ('service', 'Exchange2010SMTPAuth'),
    ('service', 'Exchange2010Protocols'),
    ('service', 'Exchange2010SMTP'),
    ('service', 'Exchange2010POP3'),
    ('service', 'Exchange2010Autodiscover'),
    ('service', 'Exchange2010Mailstore'),

    ('service', 'Exchange2013OAB'),
    ('service', 'Exchange2013IMAP4'),
    ('service', 'Exchange2013SMTPAuth'),
    ('service', 'Exchange2013Protocols'),
    ('service', 'Exchange2013SMTP'),
    ('service', 'Exchange2013POP3'),
    ('service', 'Exchange2013Autodiscover'),
    ('service', 'Exchange2013Mailstore'),

    ('service', 'Exchange2016OAB'),
    ('service', 'Exchange2016IMAP4'),
    ('service', 'Exchange2016SMTPAuth'),
    ('service', 'Exchange2016Protocols'),
    ('service', 'Exchange2016SMTP'),
    ('service', 'Exchange2016POP3'),
    ('service', 'Exchange2016Autodiscover'),
    ('service', 'Exchange2016Mailstore'),
]

EXCHANGE_MN_PACKAGES = [
    ('other', 'PrivilegesExchange'),
    ('other', 'ExchangeApplication'),
]

EXCHANGE_SN_PACKAGES = EXCHANGE_SERVICES + [
    ('other', 'MPFProviderClients'),
    ('other', '.NET2'),
]

EXCHANGE_PACKAGES_TO_REMOVE_FROM_REPO = EXCHANGE_MN_PACKAGES + EXCHANGE_SN_PACKAGES + [
    ('other', 'ExchangeMigration')
]

EXCHANGE_SCS = [
    'Exchange'
]

NOTIFICATION_TEMPLATE_LIST = [
    ["Subscription Notification", "Welcome e-mail - Hosted Exchange", "en"]
]

EXCHANGE_MODULE_NAME = 'Messaging and Collaboration'

help_link = 'https://kb.cloudblue.com/en/133390'


class CleanUpHostedExchange(CleanUpSHBase):
    def __init__(self):
        list_of_available_steps = [
            CleanUpSteps.DEACTIVATE_SERVICE_TEMPLATES,
            CleanUpSteps.CLEAR_SERVICE_NODES,
            CleanUpSteps.UNINSTALL_MODULE,
        ]
        if is_billing_installed():
            list_of_available_steps.insert(1, CleanUpSteps.SYNCHRONIZE_ONLINE_STORES)

        super(CleanUpHostedExchange, self).__init__(
            name=OAModule.EXCHANGE,
            list_of_available_steps=list_of_available_steps,
            indicator_services=EXCHANGE_SERVICES,
            services_to_detach=EXCHANGE_SERVICES,
            module_services=EXCHANGE_SN_PACKAGES,
            help_link=help_link,
        )
        self.api = openapi.OpenAPI()
        if is_billing_installed():
            self.pba_api = get_billing_api()

    def _remove_packages_from_repo(self, sc_list):
        count_packages = packages.remove_packages_from_repo(EXCHANGE_PACKAGES_TO_REMOVE_FROM_REPO)
        scs = [('sc', sc) for sc in sc_list]
        count_packages += packages.remove_packages_from_repo(scs, None)
        return count_packages

    def _uninstall_module(self):
        scs_to_rm = EXCHANGE_SCS
        self._wait_for_tasks_to_complete()
        self._remove_aps_application_resource_types('http://www.odin.com/exchange')
        self._remove_resource_types(scs_to_rm)
        self._remove_resource_classes(scs_to_rm)

        self._remove_periodic_tasks(['monitorExchangeHandler', 'monitorExchangeAddonsHandler', 'createSPLAReportHandler'])
        self._remove_sc_packages_from_mn(scs_to_rm)
        self._drop_scs_tables_seqs()
        self._reset_endpoint_to_fake_uri()
        self._redeploy_war()
        packages.remove_packages_on_hosts([1], EXCHANGE_MN_PACKAGES)

        self._unregister_module(EXCHANGE_MODULE_NAME)
        self._remove_packages(scs_to_rm)
        self._remove_exchange_messages()
        self.drop_notification_templates(NOTIFICATION_TEMPLATE_LIST)
        hosts.restart_pa_agent()
        hosts.restart_mn()
        hosts.restart_ui_hosts()

    def _wait_for_tasks_to_complete(self):
        logger.info("Waiting for tasks to complete")
        uTasks.waitTasksComplete("t.method = 'removeBrandingHandler'")

    def _reset_endpoint_to_fake_uri(self):
        con = uSysDB.connect()
        cur = con.cursor()
        cur.execute("update aps_application_instances set uri = 'https://127.0.0.1:8081/pa/rest/fake' "
                    "where app_id in (select app_id from aps_applications where name = 'ExchangeApplication')")
        cur.execute("update aps_application set endpoint_uri = 'https://127.0.0.1:8081/pa/rest/fake' "
                    "where package_id in (select id from aps_package where name = 'ExchangeApplication')")
        con.commit()

    def _redeploy_war(self):
        jboss = uPEM.JBoss()
        jboss.execCLI("-c", "deploy /usr/local/pem/ExchangeApplication/exchange-war.war --force")

    def _drop_scs_tables_seqs(self):
        con = uSysDB.connect()
        cur = con.cursor()
        self._unregister_exchange_domain_hosting(cur)
        self._drop_seqs_deps(['exch_contacts_addresses_seq',
                              'exch_mailbox_stores_seq',
                              'exch_public_folders_seq',
                              'exch_spla_sku_seq',
                              'exch_disclaimers_seq',
                              'exch_mailgroups_addresses_seq',
                              'exch_public_stores_seq',
                              'exch_temp_seq',
                              'exch_mailboxes_addresses_seq',
                              'exch_nlb_clusters_seq',
                              'exch_spla_features_seq',
                              'exch_mailbox_plans_seq',
                              'exch_public_folders_addresses_seq',
                              'exch_spla_reports_seq',
                              ],
                             cur)
        con.commit()

    def _unregister_exchange_domain_hosting(self, cur):
        cur.execute("DELETE FROM domain_hosting_configurators WHERE dhtype = 'Exchange'")
        cur.execute("DELETE FROM domain_hosting_type_categories WHERE dhtype = 'Exchange'")
        cur.execute("DELETE FROM domain_hosting_types WHERE dhtype = 'Exchange'")

    def _remove_exchange_messages(self):
        logger.info('Removing %s notification templates...', self.name)
        con = uSysDB.connect()
        cur = con.cursor()
        cur.execute("DELETE FROM message_types WHERE name IN "
                    "   ('Custom Primary email confirmation', 'Warning size limit of Exchange store is reached')")
        cur.execute("DELETE FROM messages WHERE message_key IN "
                    "   ('The warning limit of the Exchange store size is reached',"
                    "       'Confirm use of your e-mail address')")
        cur.execute("DELETE FROM messages WHERE message_key LIKE '%The reserved size of the Exchange store%'")
        cur.execute("DELETE FROM messages WHERE message_key LIKE '%custom e-mail address for your Exchange mailbox%'")
        con.commit()
