# (C)2019 Ingram Micro Inc. all Rights Reserved
import logging

from collections import defaultdict

from oacleanup.settings import OAModule, CleanUpSteps, LOG_FILE_NAME
from oacleanup.utils import resources
from oacleanup.utils.billing import is_billing_installed, get_billing_api
from oacleanup.utils.errors import CleanUpException
from poaupdater import uSysDB, openapi, apsapi
from sh_base import CleanUpSHBase

logger = logging.getLogger()

HELP_LINK = 'https://kb.cloudblue.com/134531'
LYNC_HP2 = 'Microsoft Lync Hosting Pack v2'
LYNC_HP2_UID = 'http://LyncHPv2/'


def sql_int_list(iterable):
    return ",".join(str(e) for e in iterable)


def split_list(source_list, portion_size):
    return [source_list[i:i+portion_size] for i in xrange(0, len(source_list), portion_size)]


def check_no_unsupported_rates(rt_ids):
    open_api = openapi.OpenAPI()
    token = open_api.pem.APS.getAccountToken(account_id=1)['aps_token']
    aapi = apsapi.API(url='https://localhost:6308')
    unsupported_resources = []
    for portion in split_list(map(str, rt_ids), 10):
        unsupported_resources += aapi.GET(
            path="/aps/2/resources?"
                 "implementing(http://www.odin.com/billing/Resource/1.0),"
                 "in(id,(%s))" % sql_int_list(portion),
            headers={"APS-Token": token}
        )

    unsupported_resources_map = {}
    for res in unsupported_resources:
        unsupported_resources_map[res.aps.id] = res

    con = uSysDB.connect()
    cur = con.cursor()
    cur.execute("SELECT DISTINCT st_id FROM st_resources WHERE rt_id IN (%s)" % sql_int_list(rt_ids))
    sts_to_check = [st_id for st_id, in cur.fetchall()]
    plans = []
    for portion in split_list(map(str, sts_to_check), 10):
        request_portion_size = 100  # retrieve via portions to not exceed response size limit
        start_pos = 0
        while True:
            new_plans = aapi.GET(
                path="/aps/2/resources?"
                     "implementing(http://www.odin.com/billing/ServicePlan/1.0),"
                     "limit(%d,%d),"
                     "select(serviceTemplate),"
                     "in(serviceTemplate.serviceTemplateId,(%s))" % (start_pos, start_pos + request_portion_size, sql_int_list(portion)),
                headers={"APS-Token": token})
            if not new_plans:
                break
            start_pos += request_portion_size
            plans += new_plans

    issues = []
    for plan in plans:
        if 'resourceRates' not in plan:
            continue
        for rate in plan.resourceRates:
            if rate.resourceId in unsupported_resources_map:
                issues.append("Service Plan '%s' has resource rate configured for resource '%s' to be removed. "
                              "Please remove resource rate." % (
                                  plan.name.en_US, unsupported_resources_map[rate.resourceId].name.en_US))
    return issues


def get_app_id(app_uid):
    con = uSysDB.connect()
    cur = con.cursor()
    cur.execute("SELECT app_id FROM aps_applications WHERE app_uid LIKE '%s%%'" % app_uid)
    apps = cur.fetchall()
    if apps:
        app_id = apps[0][0]
        return app_id


def get_rt_ids(app_id):
    con = uSysDB.connect()
    cur = con.cursor()
    cur.execute("""
        SELECT rtap.rt_id FROM rt_act_params rtap
            JOIN resource_types rt ON rtap.rt_id = rt.rt_id
            JOIN resource_classes rc ON rt.class_id = rc.class_id
        WHERE rc.name LIKE 'rc.saas%%' AND rtap.name = 'app_id' AND value = '%s'
    """ % app_id)
    return [int(r[0]) for r in cur.fetchall()]


def get_st_ids_by_rt_ids(rt_ids):
    if not rt_ids:
        return []
    con = uSysDB.connect()
    cur = con.cursor()
    cur.execute("""
        SELECT DISTINCT str.st_id FROM st_resources str JOIN service_templates st ON st.st_id = str.st_id
        WHERE st.for_sale = 'y' AND rt_id IN (%s)
    """ % sql_int_list(rt_ids))
    return [st_id for st_id, in cur.fetchall()]


def remove_app_instances(app_id):
    api = openapi.OpenAPI()

    result = api.pem.APS.getApplicationInstances(app_id=app_id)

    if not len(result):
        logger.info("No instances found for %s with ID #%s" % (LYNC_HP2, app_id))
        return

    for instance in result:
        request_id = api.beginRequest()
        api.pem.APS.unprovideApplicationInstance(application_instance_id=instance['application_instance_id'])
        api.commit()
        openapi.waitRequestComplete(request_id, "unprovideApplicationInstance", True)


def remove_app(app_id):
    api = openapi.OpenAPI()

    try:
        api.pem.APS.removeApplication(application_id=app_id)
    except openapi.OpenAPIError, e:
        if e.module_id == "APSManager" and e.extype_id == 7:
            logger.info("No %s found with ID #%s" % (LYNC_HP2, app_id))
        else:
            raise


class CleanUpLync(CleanUpSHBase):
    def __init__(self):
        list_of_available_steps = [
            CleanUpSteps.DEACTIVATE_SERVICE_TEMPLATES,
            CleanUpSteps.REMOVE_PACKAGES
        ]
        billing_installed = is_billing_installed()
        if billing_installed:
            list_of_available_steps.insert(1, CleanUpSteps.SYNCHRONIZE_ONLINE_STORES)

        super(CleanUpLync, self).__init__(
            name=OAModule.LYNC,
            list_of_available_steps=list_of_available_steps,
            indicator_services=[],
            services_to_detach=[],
            module_services=[],
            help_link=HELP_LINK,
        )
        self.api = openapi.OpenAPI()
        self.app_id = get_app_id(LYNC_HP2_UID)
        if billing_installed:
            self.pba_api = get_billing_api()

    def _deactivate_service_templates(self) :
        if not self._is_app_installed():
            return

        logger.info('Deactivating %s service templates...', self.name)
        st_ids = get_st_ids_by_rt_ids(get_rt_ids(self.app_id))
        if not st_ids:
            logger.info('Active service templates related to %s were not found.', self.name)
            return

        for st_id in st_ids:
            try:
                logger.debug('Deactivating the service template with ID %s...', st_id)
                self.api.pem.deactivateST(st_id=st_id)
                logger.debug('The service template with ID %s was deactivated.', st_id)
            except Exception as e:
                raise CleanUpException(
                    'Unable to deactivate the service template %s: %s. Try to deactivate it from the Provider Panel.'
                    % (st_id, e)
                )

        logger.info('%s service templates were deactivated.' % len(st_ids))

    def _is_service_templates_deactivated(self):
        issues = []
        if not self._is_app_installed():
            return True, issues

        rt_ids = get_rt_ids(self.app_id)
        issues += check_no_unsupported_rates(rt_ids)
        if issues:
            return False, issues

        st_ids = get_st_ids_by_rt_ids(rt_ids)
        if st_ids:
            logger.error(
                'There are active service templates related to the %s module. To continue with the cleanup process, '
                "deactivate them first by using the cleanup tool's command deactivate-service-templates. A list of "
                'identifiers of these service templates is located in the file %s.' % (self.name, LOG_FILE_NAME)
            )
            issues.append(
                'These are the identifiers of the service templates that you must deactivate: %s.' %
                ', '.join([str(st_id) for st_id in st_ids])
            )

        return not bool(st_ids), issues

    def _show_service_templates(self):
        st_ids = get_st_ids_by_rt_ids(get_rt_ids(self.app_id))
        if not st_ids:
            logger.info('Active service templates related to %s were not found.', self.name)
        else:
            logger.info(
                'There are active service templates related to the %s module. A list of identifiers of these service '
                'templates is located in the file %s.' % (self.name, LOG_FILE_NAME)
            )
            logger.debug(
                'These are the identifiers of the active service templates related to the %s module: %s.' %
                (self.name, ', '.join([str(st_id) for st_id in st_ids]))
            )

        return True, []

    def _remove_packages_from_repo(self, sc_list):
        # just empty implementation of abstract method
        # nothing to do here
        # all is done in overriding _remove_packages()
        pass

    def _remove_packages(self, sc_list=None):
        if not self._is_app_installed():
            return

        self._remove_aps_resource_types()

        # remove redundant service instances (if any left after customer subscriptions removal)
        con = uSysDB.connect()
        cur = con.cursor()
        si_ids_query = """
            SELECT service_instance_id FROM aps_service_instances WHERE
                service_id != 'lyncIntegration' AND app_instance_id IN
                    (SELECT app_instance_id FROM aps_application_instances WHERE app_id = %d)""" % self.app_id
        cur.execute("""
            DELETE FROM aps_service_instance_resources WHERE service_instance_id IN (%(si_ids_query)s);
            DELETE FROM aps_service_instances WHERE service_instance_id IN (%(si_ids_query)s);
            """ % {'si_ids_query': si_ids_query})
        con.commit()

        logger.info('Removing %s package...', self.name)
        remove_app_instances(self.app_id)
        remove_app(self.app_id)
        logger.info("The %s module was removed." % self.name)

    def _is_subscriptions_destroyed(self):
        issues = []
        con = uSysDB.connect()
        cur = con.cursor()
        cur.execute("""
            SELECT s.sub_id, name FROM subs_resources sr
                JOIN subscriptions s ON s.sub_id = sr.sub_id
                JOIN accounts a ON s.owner_id = a.account_id
            WHERE a.account_type = 3 AND
                rt_id IN (SELECT rt_id FROM rt_act_params
                            WHERE name = 'service_type' AND value LIKE 'http://www.parallels.com/LyncHPv2/organization%')
        """)
        subscriptions = cur.fetchall()
        if subscriptions:
            msg = 'There are %d subscriptions (both enabled and disabled) using %s application:\n    ' % (len(subscriptions), LYNC_HP2)
            msg += '\n    '.join(['Subscription ID: %d, Name: %s' % (sub_id, name) for sub_id, name in subscriptions])
            issues.append(msg)

        return not bool(subscriptions), issues

    def _remove_aps_resource_types(self):
        logger.info('Removing %s APS resources types', self.name)
        rts = get_rt_ids(self.app_id)
        for rt_id in rts:
            if not rt_id:
                continue
            logger.debug('Removing the resource type with ID %s', rt_id)
            resources.remove_resource_type(rt_id)

    def _is_app_installed(self):
        if not self.app_id:
            logger.info("There is no %s application installed. Nothing to do." % LYNC_HP2)
            return False
        else:
            return bool(self.app_id)

