# (C)2018 Ingram Micro Inc. all Rights Reserved
import logging

from oacleanup.settings import OAModule, CleanUpSteps, LOG_FILE_NAME
from oacleanup.utils import packages, hosts
from oacleanup.utils.billing import is_billing_installed, get_billing_api
from poaupdater import uSysDB, openapi
from sh_base import CleanUpSHBase

logger = logging.getLogger()

QMAIL_SERVICES = [
    ('service', 'SpamAssassin'),
    ('service', 'DrWebFilter'),
    ('service', 'drweb_daemon'),
    ('service', 'clustered-qmail'),
    ('other', 'imp'),
    ('other', 'atmail'),
    ('other', 'webmail_login'),
    ('other', 'webmail-config'),
    ('other', 'mail_schema'),
]

QMAIL_SN_PACKAGES = QMAIL_SERVICES + [
    ('other', 'SpamAssassin'),
    ('other', 'SpamAssassin-proxy'),
    ('other', 'DrWebFilter'),
    ('other', 'majordomo'),
    ('other', 'majordomo-ja'),
    ('other', 'dovecot'),
    ('other', 'CourierIMAP'),
    ('other', 'whosond'),
    ('other', 'autoresponder2'),
    ('other', 'openldap_server'),
]

NON_QMAIL_SN_PACKAGES = [
    ('service', 'MySQL5'),
    ('service', 'logparser'),
    ('service', 'logrotate'),
    ('other', 'shared_ip'),
    ('other', 'db_client'),
    ('other', 'hs_common'),
    ('other', 'service_config'),
    ('other', 'os_config'),
]


QMAIL_PACKAGES_TO_REMOVE_FROM_REPO = QMAIL_SN_PACKAGES

QMAIL_SCS = [
    'drweb_daemon',
    'drweb_filter',
    'SpamAssassin',
    'QmailCluster'
]

QMAIL_MODULE_NAME = 'Linux Mail'

help_link = 'https://kb.cloudblue.com/en/133465'


class CleanUpQmail(CleanUpSHBase):
    def __init__(self):
        list_of_available_steps = [
            CleanUpSteps.DEACTIVATE_SERVICE_TEMPLATES,
            CleanUpSteps.CLEAR_SERVICE_NODES,
            CleanUpSteps.UNINSTALL_MODULE,
        ]
        billing_installed = is_billing_installed()
        if billing_installed:
            list_of_available_steps.insert(1, CleanUpSteps.SYNCHRONIZE_ONLINE_STORES)

        super(CleanUpQmail, self).__init__(
            name=OAModule.QMAIL,
            list_of_available_steps=list_of_available_steps,
            indicator_services=QMAIL_SERVICES,
            services_to_detach=QMAIL_SERVICES,
            module_services=QMAIL_SN_PACKAGES + NON_QMAIL_SN_PACKAGES,
            help_link=help_link,
        )
        self.api = openapi.OpenAPI()
        if billing_installed:
            self.pba_api = get_billing_api()

    def _remove_packages_from_repo(self, sc_list):
        count_packages = packages.remove_packages_from_repo(QMAIL_PACKAGES_TO_REMOVE_FROM_REPO)
        scs = [('sc', sc) for sc in sc_list]
        count_packages += packages.remove_packages_from_repo(scs)
        return count_packages

    def _uninstall_module(self):
        scs_to_rm = QMAIL_SCS
        self._remove_resource_types(scs_to_rm)
        self._remove_resource_classes(scs_to_rm)

        self._remove_periodic_tasks(['taskDUCollectorsMonitor'])
        self._drop_scs_tables()
        self._remove_sc_packages_from_mn(scs_to_rm)
        self._drop_scs_tables_seqs()

        self._unregister_module(QMAIL_MODULE_NAME)
        self._remove_packages(scs_to_rm)
        self._remove_qmail_messages()

        if not self.is_module_exists("Linux Shared Hosting"):
            self.remove_common_scs()
        hosts.restart_mn()  # restart PAU to prevent unforeseen error in PCP on domain dashboard (Qmail SC is missed)
        hosts.restart_ui_hosts()

    def _drop_scs_tables_seqs(self):
        con = uSysDB.connect()
        cur = con.cursor()
        self._drop_seqs_deps(['sa_email_patterns_seq',
                              'cqmail_ldap_servers_seq',
                              'cqmail_ldap_replicas_seq',
                              'cqmail_configurations_seq',
                              'cqmail_webmails_seq',
                              'cqmail_hostings_seq',
                              'cqmail_domains_seq',
                              'cqmail_emails_seq',
                              'cqmail_mailnames_seq',
                              'cqmail_mailboxes_seq',
                              'cqmail_forwardings_seq',
                              'cqmail_filters_seq',
                              'cqmail_autoresponders_seq',
                              'cqmail_maillists_seq',
                              'cqmail_maillist_param_types_seq',
                              'cqmail_maillist_owners_seq',
                              'cqmail_maillist_approval_seq',
                              'cqmail_subscribers_seq',
                              'cqmail_ironport_operations_seq',
                              'clusters_seq',
                              'cluster_nodes_seq',
                              'lb_configuration_types_seq',
                              'load_balancers_seq',
                              'lb_serviced_protocols_seq',
                              'lb_supported_algorithms_seq',
                              'load_balancing_rules_seq',
                             ],
                             cur)
        con.commit()

    def _remove_packages_from_service_nodes_steps(self):
        hosts_count = super(CleanUpQmail, self)._remove_packages_from_service_nodes_steps()
        self._truncate_cluster_tables()
        return hosts_count

    def _truncate_cluster_tables(self):
        con = uSysDB.connect()
        cur = con.cursor()
        cur.execute('TRUNCATE TABLE clusters CASCADE')
        cur.execute('TRUNCATE TABLE cluster_nodes CASCADE')
        cur.execute('TRUNCATE TABLE lb_rule_properties CASCADE')
        cur.execute('TRUNCATE TABLE load_balancing_rules CASCADE')
        cur.execute('TRUNCATE TABLE load_balancers CASCADE')
        con.commit()

    def _drop_scs_tables(self):
        # need drop a part of tables, otherwise QmailCluster SC removal hangs up
        con = uSysDB.connect()
        cur = con.cursor()
        self._unregister_qmail_domain_hosting(cur)
        self._drop_tables_deps(['sa_services', 'sa_mailboxes', 'sa_email_patterns',
                                'drweb_filter_props',
                                'cluster_nodes', 'cluster_types', 'clusters',
                                'load_balancers', 'lb_configuration_types', 'lb_serviced_protocols',
                                'lb_supported_algorithms', 'load_balancing_rules',
                                'cqmail_ldap_servers', 'cqmail_configurations', 'cqmail_webmails', 'cqmail_mailnames',
                                'cqmail_hostings', 'cqmail_domains', 'cqmail_emails',
                                'cqmail_maillists', 'cqmail_maillist_param_types',
                                ],
                               cur)
        self._create_indep_views(['v_clusters', 'v_load_balancers', 'v_cluster_nodes', 'v_clustered_hosts',
                                  'v_load_balancing_rules', 'v_load_balancing_rules',
                                  ],
                                 cur)
        con.commit()

    def _unregister_qmail_domain_hosting(self, cur):
        cur.execute("DELETE FROM domain_hosting_configurators WHERE dhtype = 'Qmail'")
        cur.execute("DELETE FROM domain_hosting_type_categories WHERE dhtype = 'Qmail'")
        cur.execute("DELETE FROM domain_hosting_types WHERE dhtype = 'Qmail'")

    def _is_service_templates_deactivated(self):
        # It's allowed to remove Qmail services/RTs while STs are active.
        # It's only expected that there are no resources rates based on the resources.
        return True, []

    def _is_subscriptions_destroyed(self):
        issues = []
        con = uSysDB.connect()
        cur = con.cursor()
        if not uSysDB.table_exist(con, 'cqmail_hostings'):
            return True, issues
        cur.execute("SELECT Column_Name FROM INFORMATION_SCHEMA.COLUMNS WHERE Table_Name = 'cqmail_hostings' AND Column_Name = 'subscription_id';")
        if not cur.fetchone():
            return True, issues
        cur.execute('SELECT subscription_id FROM cqmail_hostings')
        subscriptions = [r[0] for r in cur.fetchall()]
        if subscriptions:
            logger.error('There are subscriptions with provisioned Qmail hosting. To continue with the cleanup process, '
                         'destroy the Qmail hosting first. The list of affected subscriptions is located in the file %s' % LOG_FILE_NAME)
            issues.append('Subscriptions including provisioned Qmail hosting: %s.' % subscriptions)
            return False, issues
        return True, issues

    def _remove_qmail_messages(self):
        logger.info('Removing %s notification templates...', self.name)
        con = uSysDB.connect()
        cur = con.cursor()
        cur.execute("DELETE FROM message_types WHERE name IN ('Mailname created', 'Mailname configuration changed', 'Test mail message')")
        cur.execute("DELETE FROM messages WHERE message_key IN ('E-mail account configuration', 'E-mail account configuration change notification', 'Test mail message for newly created mailname')")
        con.commit()
