# (C)2018 Ingram Micro Inc. all Rights Reserve
import logging

from poaupdater import openapi, uBilling, uDialog, uSysDB

from oacleanup.cleanup.base import CleanUpProcedureBase
from oacleanup.settings import LOG_FILE_NAME, OAModule, OA_NAME
from oacleanup.utils import hosts, nodes, service_templates, subscriptions, resources, packages, modules, tasks
from oacleanup.utils.errors import CleanUpException
from oacleanup.utils.billing import get_raw_billing_api

logger = logging.getLogger()

COMMON_SC_LIST_EXCEPT = [
    'MySQL',
    'LogManager',
    'logrotate',
    'DatabaseManager'
]

COMMON_SC_LIST = ['BackupManager'] + COMMON_SC_LIST_EXCEPT;

COMMON_MN_PACKAGES = [
    ('other', 'PrivilegesBackups'),
    ('other', 'PrivilegesDatabases'),
    ('other', 'PrivilegesMail')
]

COMMON_PACKAGES_TO_REMOVE_FROM_REPO = [
    ('service', 'MySQL5'),
    ('service', 'MariaDB10.1'),
    ('service', 'PgSQL'),
    ('service', 'MySQL5plus'),
    ('service', 'MySQL57'),
    ('other', 'domain_parking_common'),
    ('service', 'logparser'),
    ('service', 'logrotate'),
    ('service', 'crontabs')
]

class CleanUpSHBase(CleanUpProcedureBase):
    def __init__(
            self,
            name,
            list_of_available_steps,
            indicator_services,
            services_to_detach,
            module_services,
            help_link
    ):
        super(CleanUpSHBase, self).__init__(
            name=name,
            list_of_available_steps=list_of_available_steps,
        )
        self.indicator_services = indicator_services
        self.module_services = module_services
        self.services_to_detach = services_to_detach
        self.help_link = help_link

    def _remove_packages_from_service_nodes(self):
        logger.info('Removing packages related to %s from service nodes...', self.name)

        hosts_with_service = hosts.get_host_ids_by_services(self.services_to_detach)
        if not hosts_with_service:
            logger.info('Service nodes related to %s were not found.', self.name)
            return

        hosts_to_disable = hosts.filter_hosts_by_status(hosts_with_service, False)

        logger.debug(
            'The following service nodes related to %s were found: %s.', self.name,
            ', '.join([str(host_id) for host_id in hosts_to_disable])
        )
        hosts.mark_hosts_physically_destroyed(hosts_to_disable, True)
        try:
            hosts_count = self._remove_packages_from_service_nodes_steps()
        except Exception, e:
            raise e
        finally:
            hosts.mark_hosts_physically_destroyed(hosts_to_disable, False)

        if hosts_count == 0:
            logger.info('Service nodes related to %s were not found.', self.name)
        else:
            logger.info('Packages related to %s were removed from %s service nodes. Now you can detach these nodes.'
                        % (self.name, hosts_count))

    def _remove_packages_from_service_nodes_steps(self):
        return nodes.remove_packages_from_service_nodes(self.indicator_services, self.module_services)

    def _is_service_nodes_cleared(self):
        host_ids = hosts.get_host_ids_by_services(self.indicator_services)
        if host_ids:
            logger.error(
                'There are packages related to the %s module on the following service nodes: %s. To continue with the '
                "cleanup process, remove the packages from these service nodes by using the cleanup tool's "
                'command remove-packages-from-service-nodes.' % (self.name, ', '.join(
                    [str(host_id) for host_id in host_ids]
                ))
            )

        return not bool(host_ids), []

    def _show_service_templates(self):
        st_ids = service_templates.get_service_template_ids_by_service_names(
            [service_name for service_type, service_name in self.indicator_services],
            True
        )
        if not st_ids:
            logger.info('Active service templates related to %s were not found.', self.name)
        else:
            logger.info(
                'There are active service templates related to the %s module. A list of identifiers of these service '
                'templates is located in the file %s.' % (self.name, LOG_FILE_NAME)
            )
            logger.debug(
                'These are the identifiers of the active service templates related to the %s module: %s.' %
                (self.name, ', '.join([str(st_id) for st_id in st_ids]))
            )

        return True, []

    def _deactivate_service_templates(self):
        logger.info('Deactivating %s service templates...', self.name)
        st_ids = service_templates.get_service_template_ids_by_service_names(
            [service_name for service_type, service_name in self.indicator_services],
            True
        )
        if not st_ids:
            logger.info('Active service templates related to %s were not found.', self.name)
            return

        open_api = openapi.OpenAPI()
        for st_id in st_ids:
            try:
                logger.debug('Deactivating the service template with ID %s...', st_id)
                open_api.pem.deactivateST(st_id=st_id)
                logger.debug('The service template with ID %s was deactivated.', st_id)
            except Exception as e:
                raise CleanUpException(
                    'Unable to deactivate the service template %s: %s. Try to deactivate it from the Provider Panel.'
                    % (st_id, e)
                )

        logger.info('%s service templates were deactivated.' % len(st_ids))

    def _synchronize_online_stores(self):
        try:
            logger.info('Synchronizing online stores...')
            uBilling.PBAConf.syncStores()
            logger.info('Online stores were synchronized.')
        except Exception as e:
            raise CleanUpException('Unable to synchronize online stores: %s' % e)

    def _is_service_templates_deactivated(self):
        issues = []
        service_template_ids = service_templates.get_service_template_ids_by_service_names(
            [service_name for service_type, service_name in self.indicator_services],
            True
        )
        if service_template_ids:
            logger.error(
                'There are active service templates related to the %s module. To continue with the cleanup process, '
                "deactivate them first by using the cleanup tool's command deactivate-service-templates. A list of "
                'identifiers of these service templates is located in the file %s.' % (self.name, LOG_FILE_NAME)
            )
            issues.append(
                'These are the identifiers of the service templates that you must deactivate: %s.' %
                ', '.join([str(service_template_id) for service_template_id in service_template_ids])
            )

        return not bool(service_template_ids), issues

    def _is_subscriptions_destroyed(self):
        issues = []
        host_ids = hosts.get_host_ids_by_services(self.indicator_services)
        if not host_ids:
            return True, issues

        reseller_subscription_ids = subscriptions.get_resellers_subscription_ids_by_hosts(host_ids)
        customer_subscription_ids = subscriptions.get_customers_subscription_ids_by_hosts(host_ids)

        if customer_subscription_ids:
            if self.name == OAModule.LSH:
                alternative_path = 'or migrate '
            else:
                alternative_path = ''
                
            logger.error(
                'We found subscriptions related to the %s module. To continue with the cleanup process, destroy %s'
                'them first. The list of identifiers of these subscriptions is located in the file %s.' % (
                    self.name, alternative_path, LOG_FILE_NAME
                )
            )
            issues.append(
                'These are the identifiers of the subscriptions that are related to the %s module: %s.' %
                (self.name, ', '.join([str(subscription_id) for subscription_id in customer_subscription_ids]))
            )

        if reseller_subscription_ids:
            logger.error(
                'There are reseller subscriptions related to the %s module. '
                'Remove all webspaces and websites related to these subscriptions. '
                'A list of identifiers of these subscriptions is located in the file %s.' %
                (self.name, LOG_FILE_NAME)
            )
            issues.append(
                'These are the identifiers of the reseller subscriptions that are related to the %s module: %s.' %
                (self.name, ', '.join([str(subscription_id) for subscription_id in reseller_subscription_ids]))
            )

        return not (bool(customer_subscription_ids) or bool(reseller_subscription_ids)), issues

    def _remove_packages(self, sc_list=None):
        logger.info('Removing %s packages from the repository...', self.name)
        count_packages = self._remove_packages_from_repo(sc_list)
        if count_packages == 0:
            logger.info('Packages related to %s were not found in the repository.' % self.name)
        else:
            logger.info('%s packages related to %s were removed from the repository.' % (count_packages, self.name))

        logger.info("The %s module was removed" % self.name)

    def _remove_resource_types(self, scs_to_remove):
        logger.info('Removing the resource types related to %s...', self.name)
        resources.remove_resource_types(scs_to_remove)
        logger.info('Resource types related to %s were removed.', self.name)

    def _remove_resource_classes(self, scs_to_remove):
        logger.info('Removing the resources classes related to %s...', self.name)
        rc_count = resources.remove_resource_classes(scs_to_remove)
        if rc_count == 0:
            logger.info('Resources classes related to %s were not found.', self.name)
        else:
            logger.info('%s resources classes related to %s were removed.' % (rc_count, self.name))

    def _remove_periodic_tasks(self, method_names):
        con = uSysDB.connect()
        cur = con.cursor()
        for method_name in method_names:
            for task in tasks.find_tasks(cur, method_name):
                task.delete(cur)
        con.commit()

    def _remove_sc_packages_from_mn(self, scs_to_remove):
        logger.info("Removing %s packages from the management node...", self.name)
        packages_count = packages.remove_sc_packages_from_mn(scs_to_remove)
        if packages_count == 0:
            logger.info('Packages related to %s were not found on the management node.' % self.name)
        else:
            logger.info('%s packages related to %s were removed from the management node.' % (packages_count, self.name))

    def drop_sc(self, sc_name):
        logger.info("Removing '%s' SC", sc_name)
        con = uSysDB.connect()
        cur = con.cursor()
        cur.execute("SELECT pkg_id FROM packages WHERE name = '%s' AND ctype = 'sc'" % sc_name)
        result = cur.fetchone()
        if not result:
            return
        for pkg_id in result:
            cur.execute("SELECT component_id FROM components WHERE pkg_id = %s" % pkg_id)
            result = cur.fetchone()
            if not result:
                continue
            comp_id = result[0]
            cur.execute("DELETE FROM component_dependencies WHERE component_id = %s" % comp_id)
            cur.execute("DELETE FROM components WHERE pkg_id = %s" % pkg_id)
        cur.execute("DELETE FROM service_classes WHERE name = '%s'" % sc_name)
        con.commit()

    def _unregister_module(self, module_name):
        logger.info("Unregistering the %s module from %s..." % (module_name, OA_NAME))
        if not modules.is_module_installed(module_name):
            logger.debug("The %s module has already been unregistered." % module_name)
        else:
            modules.unregister_modules(module_name)
            logger.info("The %s module was unregistered from %s." % (module_name, OA_NAME))

    def _remove_packages_from_repo(self, sc_list):
        raise NotImplementedError()

    def _drop_tables_deps(self, names, dbcur):
        for name in names:
            self._drop_table_deps(name, dbcur)

    def _drop_table_deps(self, name, dbcur):
        # drop table to w/a incorrect order of tables removal during SCs de-installing
        dbcur.execute('DROP TABLE IF EXISTS ' + name + ' CASCADE')
        # create fake table to prevent SC deinstalling failure due to table absence
        dbcur.execute('CREATE TABLE ' + name + ' (id INT)')

    def _create_indep_views(self, views, dbcur):
        for view in views:
            self._create_indep_view(view, dbcur)

    def _create_indep_view(self, view, dbcur):
        # create fake view to prevent SC deinstalling failure due to view absence
        dbcur.execute('CREATE OR REPLACE VIEW ' + view + ' AS SELECT 1')

    def _drop_seqs_deps(self, names, dbcur):
        for name in names:
            self._drop_seq_deps(name, dbcur)

    def _drop_seq_deps(self, name, dbcur):
        dbcur.execute('DROP SEQUENCE IF EXISTS ' + name + ' CASCADE')

    def _unprovision_resources(self, subscription_ids):
        subscription_ids = [int(x) for x in subscription_ids.split(',')]
        logger.info('Unprovisioning %s resources from the specified subscriptions...' % self.name)
        resources_to_unprovision_by_sub = {}
        resources_to_unprovision = set()
        for sub_id, rt_id, restype_name, rt_attr in resources.get_resource_types_for_services_by_subscriptions(
            [sc_name for _, sc_name in self.indicator_services], subscription_ids
        ):
            if sub_id not in resources_to_unprovision_by_sub:
                resources_to_unprovision_by_sub[sub_id] = set()

            resources_to_unprovision_by_sub[sub_id].add((rt_id, restype_name, rt_attr))
            resources_to_unprovision.add((rt_id, restype_name, rt_attr))

        if not resources_to_unprovision:
            logger.info('There are no resources to unprovision.')
            return

        logger.info('The following resources were detected:')
        for rt_id, restype_name, rt_attr in resources_to_unprovision:
            logger.info('\t{rt_id}: {restype_name} ({rt_attr})'.format(
                rt_id=rt_id, restype_name=restype_name,
                rt_attr='with the "{rt_attr}" provisioning attribute'.format(rt_attr=rt_attr) if rt_attr
                else 'without provisioning attributes'
            ))

        answer = uDialog.askYesNo(
            "Please note that if databases were created for websites or applications running on {module}, they will "
            "be unprovisioned too. Are you sure that you want to unprovision these resources?".format(module=self.name)
        )
        if not answer:
            logger.info('The unprovisioning was cancelled.')
            return

        subscription_with_databases = subscriptions.get_subscriptions_with_databases(subscription_ids)
        if len(subscription_with_databases) > 0:
            logger.info('The following subscriptions with databases were detected: %s' % ', '.join(
                [str(x) for x in subscription_with_databases]
            ))
            answer = uDialog.askVariants(
                'If the databases were created for websites or applications running on  {module}, they will be '
                'unprovisioned. Do you want to unlink databases from these subscriptions?'.format(module=self.name),
                variants=['Yes', 'No', 'Abort']
            )
            if answer == 'Abort':
                return
            elif answer == 'Yes':
                subscriptions.unlink_databases_from_subscriptions(subscription_ids)

        self._unlink_aps_application_from_subscriptions(resources_to_unprovision_by_sub.keys())

        open_api = openapi.OpenAPI()
        for subscription_id in resources_to_unprovision_by_sub.keys():
            logger.debug('Unprovisioning %s resources from the subscriptions %s...' % (self.name, subscription_id))
            subscription_resources_to_unprovision = [
                {
                    'resource_type_id': rt_id,
                    'limit': 0
                } for rt_id in sorted(set([x for x, _, _ in resources_to_unprovision_by_sub[subscription_id]]))

            ]

            if subscription_resources_to_unprovision:
                open_api.pem.setResourceTypeLimits(
                    subscription_id=subscription_id,
                    autoprovide=True,
                    resource_limits=subscription_resources_to_unprovision
                )

        logger.info('All resources related to %s were unprovisioned from the specified subscriptions.' % self.name)

    def _remove_aps_application_resource_types(self, app_uid):
        con = uSysDB.connect()
        cur = con.cursor()
        cur.execute("SELECT rp.rt_id FROM aps_applications ap "
                    "JOIN rt_act_params rp ON (ap.app_id::varchar(255) = rp.value AND rp.name = 'app_id') "
                    "JOIN resource_types rt ON (rp.rt_id = rt.rt_id) "
                    "JOIN resource_classes rc ON (rc.class_id = rt.class_id AND rc.name LIKE 'rc.saas%%') "
                    "WHERE ap.app_uid = '%s';" % app_uid)
        rts = [int(r[0]) for r in cur.fetchall()]
        logger.info('Removing %s APS resources types', self.name)
        for rt_id in rts:
            if not rt_id:
                continue
            logger.debug('Removing the resource type with ID %s', rt_id)
            resources.remove_resource_type(rt_id)

    def remove_common_scs(self):
        #Method drops common Linux Shared's and Linux Mail's scs
        self._remove_resource_types(COMMON_SC_LIST)
        self._remove_resource_classes(COMMON_SC_LIST)
        packages.remove_packages_on_hosts([1], COMMON_MN_PACKAGES)
        self.drop_sc('BackupManager')
        self._remove_sc_packages_from_mn(COMMON_SC_LIST_EXCEPT)
        self._drop_common_scs_tables()
        self._drop_common_scs_seqs()
        self.remove_common_packages_from_repo()

    def _drop_common_scs_tables(self):
        con = uSysDB.connect()
        cur = con.cursor()
        tables = ['backups_groups',
                  'backups_backups',
                  'backups_restore_data',
                  'backups_restore_files',
                  'backups_configuration'
                  ]
        for table in tables:
            cur.execute("DROP TABLE IF EXISTS %s CASCADE" % table)
        con.commit()

    def _drop_common_scs_seqs(self):
        con = uSysDB.connect()
        cur = con.cursor()
        self._drop_seqs_deps(['backups_groups_seq'], cur)
        con.commit()

    def is_module_exists(self, module_name):
        con = uSysDB.connect()
        cur = con.cursor()
        cur.execute("SELECT 1 FROM modules WHERE name = '%s' AND status != 'n'" % module_name)
        return cur.fetchall()

    def remove_common_packages_from_repo(self):
        packages.remove_packages_from_repo(COMMON_MN_PACKAGES + COMMON_PACKAGES_TO_REMOVE_FROM_REPO)
        scs = [('sc', sc) for sc in COMMON_SC_LIST]
        packages.remove_packages_from_repo(scs)

    def drop_notification_templates(self, template_list):
        raw_pba_api = get_raw_billing_api()
        all_template_list = raw_pba_api.MESSAGE.GetMyTemplateList()['Result'][0]
        logger.debug('Deleting the following notification templates: %s', str(template_list))
        for template in template_list:
            msg_class_name = template[0]
            tmpl_name = template[1]
            locale_id = template[2]
            if self._is_template_exists(msg_class_name, tmpl_name, locale_id, all_template_list):
                try:
                    raw_pba_api.MESSAGE.TemplateMoveToArc(msg_class_name, tmpl_name, locale_id)
                except Exception as e:
                    logger.info("Unable to delete Notification Template '%s'. Error message: %s" % (tmpl_name, (str(e))))
                    raise e

    def _is_template_exists(self, msg_class_name, tmpl_name, locale_id, all_template_list):
        for template in all_template_list:
            curr_msg_class_name = template[1]
            curr_tmpl_name = template[2]
            curr_locale_id = template[3]
            if curr_msg_class_name == msg_class_name and curr_tmpl_name == tmpl_name and curr_locale_id == locale_id:
                return True
        return False
