# (C)2021 Ingram Micro Inc. all Rights Reserved
from sh_base import CleanUpSHBase

from oacleanup.settings import OAModule, CleanUpSteps
from oacleanup.utils import packages, nodes, hosts

from poaupdater import uSysDB, openapi, uPEM, uHCL, uTasks

import os
import sys
import logging

logger = logging.getLogger()

WHI_SERVICES = [
    ('service', 'ADLockoutMonitor'),
    ('service', 'ADDomainController'),
    ('service', 'WindowsProvisioningEngine')
]

WHI_MN_PACKAGES = [
    ('other', 'PrivilegesAD'),
]

WHI_SN_PACKAGES =  [
    ('other', 'IISCertObj_Hotfix'),
    ('other', 'SQL_Management_Objects'),
    ('other', '.NET45'),
    ('other', '.NET4')

] + WHI_SERVICES

PACKAGES_TO_REMOVE_FROM_REPO = WHI_MN_PACKAGES + WHI_SN_PACKAGES

WHI_SCS = [
    'ActiveDirectory',
    'WinHostControl'
]

help_link = 'https://cloudblue.freshdesk.com/support/solutions/articles/44002201597'

class CleanUpWHI(CleanUpSHBase):
    def __init__(self):
        list_of_available_steps = [
            CleanUpSteps.CLEAR_SERVICE_NODES,
            CleanUpSteps.UNINSTALL_MODULE,
        ]

        super(CleanUpWHI, self).__init__(
            name=OAModule.WHI,
            list_of_available_steps=list_of_available_steps,
            indicator_services=WHI_SERVICES,
            services_to_detach=WHI_SERVICES,
            module_services=WHI_SN_PACKAGES,
            help_link=help_link,
        )
        self.api = openapi.OpenAPI()

    def _remove_packages_from_repo(self, sc_list):
        count_packages = packages.remove_packages_from_repo(PACKAGES_TO_REMOVE_FROM_REPO)
        scs = [('sc', sc) for sc in sc_list]
        count_packages += packages.remove_packages_from_repo(scs, None)
        return count_packages

    def _uninstall_module(self):
        scs_to_rm = WHI_SCS
        self._mark_hosts_as_destroyed()
        self._remove_ui_props_files()
        self._remove_resource_types(scs_to_rm)
        self._remove_resource_classes(scs_to_rm)

        self._remove_periodic_tasks(['monitorPrefDcHandler'])
        packages.remove_packages_on_hosts([1], WHI_MN_PACKAGES)
        self.drop_sc('ActiveDirectory')
        self._remove_sc_packages_from_mn(scs_to_rm)
        self._drop_scs_tables_seqs()

        self._remove_packages(scs_to_rm)
        hosts.restart_pa_agent()
        hosts.restart_mn()
        hosts.restart_ui_hosts()

    def _drop_scs_tables_seqs(self):
        con = uSysDB.connect()
        cur = con.cursor()

        self._drop_seqs_deps(['actdir_organization_units_seq',
                              'actdir_entities_seq',
                              'actdir_ext_services_seq',
                              'actdir_sync_sessions_seq',
                              'actdir_sync_items_seq',
                              'actdir_sync_settings_seq'],
                             cur)

        cur.execute("DROP VIEW IF EXISTS actdir_v_subscriptions_ou CASCADE")
        TABLES = ['actdir_account_wide_subscriptions',
                  'actdir_contact_props',
                  'actdir_customer_units',
                  'actdir_entities',
                  'actdir_entity_types',
                  'actdir_ext_services',
                  'actdir_group_membership',
                  'actdir_group_membership_removing',
                  'actdir_group_props',
                  'actdir_hosts',
                  'actdir_hs_users',
                  'actdir_linked_sids',
                  'actdir_organization_units',
                  'actdir_reseller_units',
                  'actdir_sync_items',
                  'actdir_sync_sessions',
                  'actdir_sync_settings',
                  'actdir_sys_cfg',
                  'actdir_tags',
                  'actdir_user_props',
                  'actdir_windows_domains',
                  'win_node_routes']
        for table in TABLES:
            cur.execute("DROP TABLE IF EXISTS %s CASCADE" % table)
        con.commit()

    def _mark_hosts_as_destroyed(self):
        active_directory_sc_is_installed = uPEM.is_sc_installed("ActiveDirectory")
        if not active_directory_sc_is_installed:
            logger.info("ActiveDirectory sc is not found, so 'Mark ActiveDirectory based hosts as destroyed' step is skipped.")
            return
        logger.info("Mark ActiveDirectory based hosts as destroyed.")
        con = uSysDB.connect()
        cur = con.cursor()

        cur.execute("SELECT host_id FROM actdir_hosts")
        hosts = cur.fetchall()
        for host in hosts:
            cur.execute("UPDATE hosts SET is_physically_destroyed = 'y' WHERE host_id = %s" % host[0])
        con.commit()

    def _remove_ui_props_files(self):
        logger.info("Remove non-actual property files from UI nodes.")
        conf_dir = self._get_conf_dir()
        self._remove_file_if_exists(conf_dir, "whc.properties")
        self._remove_file_if_exists(conf_dir, "exchange.properties")
        self._remove_file_if_exists(conf_dir, "active_directory.properties")

    def _get_conf_dir(self):
        osa_path = uPEM.getMNInfo()[1]
        sys.path.append('/'.join([osa_path, 'u']))
        import bootstrap
        bootstrap.setJavaHome()
        return os.path.join(bootstrap.getJBossDir(osa_path), 'puitconf')

    def _remove_file_if_exists(self, folder_path, file_name):
        file_path = os.path.join(folder_path, file_name)
        for host_id in uPEM.getAllUiHosts():
            request = uHCL.Request(host_id, user='root', group='root')
            logger.info('Delete file %s on host %s.' % (file_path, host_id))
            request.command("rm -f %s" % file_path)
            request.perform()

    def _remove_packages_from_service_nodes_steps(self):
        host_count = 0;
        for service in self.indicator_services:
            host_count += nodes.remove_packages_from_service_nodes([service], self.module_services)
            uTasks.waitTasksComplete("t.status = 'u' and t.status = 'e'")
        return host_count;