# (C)2018 Ingram Micro Inc. all Rights Reserve
from poaupdater import uPackaging, uSysDB

from oacleanup.utils.errors import CleanUpException
from oacleanup.utils.hosts import mark_hosts_physically_destroyed
from oacleanup.utils.modules import is_module_installed
import nodes


import logging

logger = logging.getLogger()


def is_package_installed_on_host(host_id, name, ctype):
    con = uSysDB.connect()
    cur = con.cursor()

    q = 'SELECT component_id ' \
        'FROM packages p ' \
        'JOIN components c ON (c.pkg_id = p.pkg_id) ' \
        "WHERE p.name=%s AND p.ctype=%s AND c.uninstalled = 'n'"
    if host_id:
        q += ' AND c.host_id = %s'
        cur.execute(q, name, ctype, host_id)
    else:
        cur.execute(q, name, ctype)

    comp = cur.fetchone()
    if not comp:
        return False

    return comp[0]


def get_all_packages_on_host(host_id):
    con = uSysDB.connect()
    cur = con.cursor()

    query = ('SELECT name FROM packages'
             ' JOIN components USING (pkg_id)'
             ' WHERE host_id = %s')
    cur.execute(query, host_id)

    return [name for name, in cur.fetchall()]


def sc_has_dependencies(sc_name, exc_packages):
    (sc_id, sc_ver) = uPackaging.findHostComponentId(nodes.MN_NODE_ID, sc_name, 'sc')
    sc_deps = set([p.name for p in uPackaging.getDependencies(sc_id)])
    if exc_packages:
        sc_deps -= set(exc_packages)

    return len(sc_deps) > 0


def scs_have_services(scs):
    con = uSysDB.connect()
    cur = con.cursor()

    query = ('SELECT service_id FROM services'
             ' JOIN service_classes USING (sc_id)'
             ' WHERE service_classes.name IN %s'
             ' LIMIT 1')
    cur.execute(query, (tuple(scs),))

    return cur.fetchone()


def get_service_packages_by_scs(scs):
    con = uSysDB.connect()
    cur = con.cursor()

    query = ('SELECT p.ctype, p.name'
             ' FROM packages p'
             ' JOIN components USING (pkg_id)'
             ' JOIN services USING (component_id)'
             ' JOIN service_classes sc USING (sc_id)'
             ' WHERE sc.name IN %s')
    cur.execute(query, (tuple(scs),))

    return [(ctype, name) for ctype, name in cur.fetchall()]


def get_service_on_host(host_id, name):
    con = uSysDB.connect()
    cur = con.cursor()

    query = ('SELECT service_id'
             ' FROM services'
             ' WHERE host_id = %s'
             ' AND name = %s')
    cur.execute(query, host_id, name)

    res = cur.fetchone()
    if not res:
        return None

    return res[0]


def get_package_props(host_id, pkg_name):
    con = uSysDB.connect()
    cur = con.cursor()

    query = ('SELECT p.name, p.value'
             ' FROM v_props p'
             ' JOIN components USING (component_id)'
             ' JOIN packages pkg USING (pkg_id)'
             ' WHERE host_id = %s'
             ' AND pkg.name = %s')
    cur.execute(query, host_id, pkg_name)

    return {name: value for name, value in cur.fetchall()}


def remove_packages_on_hosts(host_ids, services):
    for host_id in host_ids:
        logger.info("Removing packages from the host #%s...", host_id)
        for ctype, name in services:
            if is_package_installed_on_host(host_id, name, ctype):
                logger.info("\t-> %s", name)
                uPackaging.removePackage(name, ctype, host_id, True)
                logger.debug('The %s package was removed from the host %s.', name, host_id)


def remove_component_self_deps(component_id):
    deps = uPackaging.getDependencies(component_id)
    for dep in deps:
        if component_id == dep.component_id:
            con = uSysDB.connect()
            cur = con.cursor()

            query = ('DELETE FROM component_dependencies'
                     ' WHERE component_id = dep_id'
                     ' AND component_id = %s')
            cur.execute(query, component_id)
            con.commit()
        else:
            remove_component_self_deps(dep.component_id)


def force_remove_packages_on_host(host_id, packages):
    con = uSysDB.connect()
    cur = con.cursor()

    for ctype, name in packages:
        component_id = is_package_installed_on_host(host_id, name, ctype)
        if component_id:
            deps = uPackaging.getDependencies(component_id)
            while deps:
                uPackaging.removeComponent(deps[0].host_id, deps[0].component_id, deps[0].hostname, deps[0].name, True)
                deps = uPackaging.getDependencies(component_id)

            query = ('DELETE FROM components c'
                     ' USING packages p'
                     ' WHERE c.pkg_id = p.pkg_id'
                     ' AND c.host_id = %s'
                     ' AND p.ctype = %s'
                     ' AND p.name = %s')
            cur.execute(query, host_id, ctype, name)

    con.commit()


def remove_sc_packages_from_mn(packages):
    packages_count = 0
    for name in packages:
        try:
            if not scs_have_services([name]) and is_package_installed_on_host(nodes.MN_NODE_ID, name, 'sc'):
                logger.info("\t-> %s", name)
                uPackaging.removePackage(name, 'sc', nodes.MN_NODE_ID, True)
                packages_count += 1
        except Exception as e:
            raise CleanUpException("Unable to remove the %s package from the management node: %s." % (name, e))

        logger.debug("The package %s was removed from the management node.", name)

    return packages_count


def remove_packages_from_repo(packages, opsyses=None):
    con = uSysDB.connect()
    cur = con.cursor()

    count_packages = 0
    for ctype, name in packages:
        query = ('SELECT pkg_id FROM packages'
                 ' JOIN platforms USING (platform_id)'
                 ' WHERE name = %s AND ctype = %s')
        if opsyses:
            query += ' AND opsys IN %s'
            cur.execute(query, (name, ctype, tuple(opsyses),))
        else:
            cur.execute(query, (name, ctype))

        for pkg_id, in cur.fetchall():
            try:
                logger.info("\t-> #%d %s", pkg_id, name)
                uPackaging.removePkgFromRepo(pkg_id, name)
                count_packages += 1
            except Exception as e:
                raise CleanUpException(
                    "Unable to remove the package '%s' from the repository: %s. "
                    'Check that this package is not installed on any hosts. '
                    'If it is installed, remove it from all hosts.' % (name, e))

            logger.debug("The package #%d '%s' was removed from the repository.", pkg_id, name)

    return count_packages


def remove_common_sc(sc_list, module_name):
    from oacleanup.cleanup.lsh import LSH_MODULE_NAME
    from oacleanup.cleanup.wsh import WSH_MODULE_NAME

    if module_name == LSH_MODULE_NAME:
        other_module = WSH_MODULE_NAME
    else:
        other_module = LSH_MODULE_NAME

    if is_module_installed(other_module):
        COMMON_SCS = [
            'WebFileManager',
            'SSLProxy',
            'InternalIPResolver',
            'DomainParking',
            'DatabaseManager',
            'ActiveDirectory',
            'WinHostControl',
            'BackupManager',
            'logrotate',
            'LogManager',
        ]
        return list(set(sc_list) - set(COMMON_SCS))
    else:
        return sc_list
