# (C)2018 Ingram Micro Inc. all Rights Reserve
from poaupdater import openapi, uResourceManagement, uSysDB

from oacleanup.utils.billing import is_billing_installed, get_billing_api
from oacleanup.utils.billing_db import get_connection_to_billing_db
from oacleanup.utils.errors import CleanUpException

import logging
import sys

logger = logging.getLogger()


def get_service_templates_by_resource(rt_id):
    """
    Returns the list of STs which contains the RT.
    STs are ordered from ones owned by deepest vendor, and provider's ST will be at the end.
    :param rt_id: RT ID
    :return: list of STs containing the given RT
    """
    con = uSysDB.connect()
    cur = con.cursor()

    # the query of unordered list of pairs (ST, parent ST); "parent" relation is in projection to given RT
    query = """
        SELECT str.st_id as st_id, src_s.st_id as src_st_id
        FROM st_resources str
        JOIN subs_resources src_sr ON src_sr.rt_instance_id=str.src_rt_instance_id
        JOIN subscriptions src_s ON src_s.sub_id=src_sr.sub_id
        WHERE str.rt_id=%s"""
    cur.execute(query, rt_id)

    def __st_sort(st_relations):
        """Sorts list of pairs (ST, parent ST); parent ST is always before its children"""
        st_ordered_list = list()
        for st_id, src_st_id in st_relations:
            for i in range(0, len(st_ordered_list) + 1):
                if i == len(st_ordered_list) or st_id == st_ordered_list[i][1]:
                    st_ordered_list.insert(i, (st_id, src_st_id))
                    break
        return st_ordered_list

    # return reversed list of STs, so the children before parent
    st_ids = [p[0] for p in reversed(__st_sort(cur.fetchall()))]
    logger.debug('RT #%s contained in STs: %s', rt_id, st_ids)
    return st_ids


def is_resource_in_service_template(st_id, rt_id):
    return st_id in get_service_templates_by_resource(rt_id)


def get_resource_types_for_sc(sc_name):
    con = uSysDB.connect()
    cur = con.cursor()

    query = ('SELECT rt_id FROM resource_types'
             ' JOIN resource_classes USING (class_id)'
             ' JOIN service_classes USING (sc_id)'
             ' WHERE service_classes.name = %s')
    cur.execute(query, sc_name)

    return [rt_id for rt_id, in cur.fetchall()]


def get_root_resource_types_for_sc(sc_name):
    con = uSysDB.connect()
    cur = con.cursor()

    query = ('SELECT rt_id FROM resource_types'
             ' JOIN resource_classes USING (class_id)'
             ' JOIN service_classes USING (sc_id)'
             ' WHERE parent_class_id IS NULL'
             ' AND service_classes.name = %s')
    cur.execute(query, sc_name)

    return [rt_id for rt_id, in cur.fetchall()]


def get_nonroot_resource_classes_for_sc(sc_name):
    con = uSysDB.connect()
    cur = con.cursor()

    query = ('SELECT rc.name FROM resource_classes rc'
             ' JOIN service_classes sc USING (sc_id)'
             ' WHERE parent_class_id IS NOT NULL'
             ' AND sc.name = %s'
             ' ORDER BY class_id DESC')
    cur.execute(query, sc_name)

    return [name for name, in cur.fetchall()]


def get_root_resource_classes_for_sc(sc_name):
    con = uSysDB.connect()
    cur = con.cursor()

    query = ('SELECT rc.name FROM resource_classes rc'
             ' JOIN service_classes sc USING (sc_id)'
             ' WHERE parent_class_id IS NULL'
             ' AND sc.name = %s'
             ' ORDER BY class_id DESC')
    cur.execute(query, sc_name)

    return [name for name, in cur.fetchall()]


def __archive_orders(pba_con, pba_api, rt_ids):
    root_rt_id = rt_ids[0]
    rt_ids_in_clause = ','.join(map(str, rt_ids))
    pba_cur = pba_con.cursor()
    archived_orders_cnt = 0
    while True:
        pba_cur.execute('''
            WITH RECURSIVE
            -- a chunk of orders linked to RTs
            -- limitations: dependencies between orders not applied
            affected_orders AS (
                SELECT DISTINCT so."OrderID"
                    FROM "ResItem" ri
                    JOIN "OItem" oi USING("OIID")
                    JOIN "SalesOrder" so ON so."OrderID"=oi."OrderDocOrderID"
                    WHERE ri."resourceID" IN ({0})
                    OR ri."to_resourceID" IN ({0})
                UNION
                SELECT DISTINCT so."OrderID"
                    FROM "OrdDet" od
                    JOIN "SalesOrder" so ON so."OrderID"=od."OrderID"
                    WHERE od."resourceID" IN ({0})
                LIMIT 100000
            ),
            -- recursive query which sorts order details according to dependencies, adding the level of dependency
            -- limitations: duplicates are possible; independent order details are last 
            orddet_dependants_rec("DetID", "OrderID", "Corrected_DetID", "Original_DetID", lvl) as (
                SELECT "DetID", "OrderID", "Corrected_DetID", "Original_DetID", 0
                    FROM "OrdDet"
                    WHERE "OrderID" IN (SELECT "OrderID" FROM affected_orders)
                UNION ALL
                SELECT od."DetID", od."OrderID", od."Corrected_DetID", od."Original_DetID", ds.lvl+1
                    FROM "OrdDet" od, orddet_dependants_rec ds
                    WHERE od."Original_DetID"=ds."DetID" OR od."Corrected_DetID"=ds."DetID"
            ),
            -- transform from order details to orders, and append orders which could have no details (so not ordered)
            orders_leveled AS (
                SELECT "OrderID", lvl FROM orddet_dependants_rec
                -- first query does not include orders not linked with order details, adding them explicitly
                UNION ALL
                SELECT "OrderID", 0 FROM affected_orders
            ),
            -- sort orders so that orders goes before its dependencies 
            orders_sorted_by_level AS (
                SELECT "OrderID", MAX(lvl)
                FROM orders_leveled
                GROUP BY "OrderID"
                ORDER BY MAX(lvl) DESC
            )
            -- the main query to go through orders according to dependencies
            SELECT so."OrderID", so."OrderStatusID"
            FROM orders_sorted_by_level
            JOIN "SalesOrder" so USING("OrderID")
            '''.format(rt_ids_in_clause))
        orders_to_archive = pba_cur.fetchall()
        if not orders_to_archive:
            break
        for order_id, order_status in orders_to_archive:
            if order_status not in ['CP', 'CL']:
                raise CleanUpException(
                    'RT #%s contained in order #%s which cannot be archived due to its non-final status %s.'
                    ' Order must be in either Completed or Cancelled status.'
                    ' Process the order manually and re-run the script.'
                    % (root_rt_id, order_id, order_status))
            logger.debug('Archiving order #%s (status %s)', order_id, order_status)
            pba_api.SalesOrderMoveToArc(order_id)
            archived_orders_cnt += 1
            if archived_orders_cnt % 100 == 0:
                logger.info('Orders archived: %s (counting progress)', archived_orders_cnt)
    if archived_orders_cnt > 0:
        logger.info('Orders archived successfully: %s', archived_orders_cnt)


def remove_resource_type(root_rt_id):
    logger.info('Removing RT #%s', root_rt_id)
    api = openapi.OpenAPI()
    con = uSysDB.connect()
    cur = con.cursor()
    cur.execute("SELECT rt_id FROM resource_types WHERE rt_id=%s OR parent_path LIKE %s ORDER BY parent_path",
                root_rt_id, 'x{}x%'.format(root_rt_id))
    rt_ids = [_rt_id for _rt_id, in cur.fetchall()]
    rt_ids_in_clause = ','.join(map(str, rt_ids))
    if not rt_ids:
        logger.info('The resource type #%s is not found', root_rt_id)
        return
    logger.debug('RTs including sub-resources for remove: %s', rt_ids)
    # mark as non-system to handle license items
    cur.execute("UPDATE resource_types rt SET system = 'n' WHERE rt_id IN ({})".format(rt_ids_in_clause))
    con.commit()
    try:
        # process RTs in billing (needed to allow removing via OpenAPI)
        if is_billing_installed():
            pba_api = get_billing_api()
            pba_con = get_connection_to_billing_db()
            pba_cur = pba_con.cursor()
            # archive orders related to RTs
            __archive_orders(pba_con, pba_api, rt_ids)
            # archive RTs in composite resources
            for res in pba_api.IncludedResourcesGetIncludedResourcesList():
                comp_id = res[0]
                inc_rt_id = res[1]
                if inc_rt_id in rt_ids:
                    logger.debug('Archiving RT #%s in the composite resource #%s', inc_rt_id, comp_id)
                    pba_api.IncludedResourcesMoveToArc(comp_id, inc_rt_id)
            # archive resource rates based on RTs
            pba_cur.execute('''
                SELECT "PlanRateID", "PlanID"
                FROM "PlanRate"
                WHERE "resourceID" IN ({})'''.format(rt_ids_in_clause))
            for rate_id, plan_id in pba_cur.fetchall():
                logger.debug('Archiving resource rate #%s in plan #%s', rate_id, plan_id)
                pba_api.PlanRateMoveToArc(rate_id)
        # exclude from STs
        for st_id in get_service_templates_by_resource(root_rt_id):
            # XXX as a workaround, remove records from subs_resources in advance,
            # otherwise in the case of a lot of subscriptions on ST, the 5 minutes timeout exceeds
            logger.debug('Removing RT #%s from subscriptions of ST #%s', root_rt_id, st_id)
            inner_con = uSysDB.connect()
            inner_cur = inner_con.cursor()
            inner_cur.execute("""
                DELETE FROM subs_resources
                WHERE rt_id IN ({0})
                AND sub_id IN (SELECT sub_id FROM subscriptions WHERE st_id={1})
            """.format(rt_ids_in_clause, st_id))
            inner_con.commit()
            logger.debug('Removing RT #%s from ST #%s', root_rt_id, st_id)
            api.pem.removeResourceTypeFromServiceTemplate(owner_id=1, st_id=st_id, rt_id=root_rt_id)
        # remove RT
        logger.debug('Removing RT #%s', root_rt_id)
        api.pem.removeResourceType(rt_id=root_rt_id)
    except CleanUpException:
        raise
    except Exception as e:
        raise CleanUpException(
            'Unable to remove the resource type #%d from service templates: %s. '
            'Try to remove it from the Provider Panel.' % (root_rt_id, e)
        ), None, sys.exc_info()[2]
    logger.info('Successfully removed RT #%s', root_rt_id)


def remove_resource_types(service_controller_list):
    for sc_name in service_controller_list:
        for root_rt_id in get_root_resource_types_for_sc(sc_name):
            logger.debug('Removing the root resource type #%s', root_rt_id)
            remove_resource_type(root_rt_id)


def remove_resource_classes(scs):
    rc_count = 0

    con = uSysDB.connect()
    api = openapi.OpenAPI()

    for sc_name in scs:
        for name in get_nonroot_resource_classes_for_sc(sc_name):
            logger.debug('Removing the non root resource class %s', name)
            uResourceManagement.deleteResourceClass(con, name)
            rc_count += 1
        con.commit()

        for name in get_root_resource_classes_for_sc(sc_name):
            logger.debug('Removing the root resource class %s', name)
            api.pem.unregisterResourceClass(name=name)
            rc_count += 1

    return rc_count


def get_resource_types_for_services_by_subscriptions(service_names, subscription_ids):
    if not service_names:
        return []

    # resource having at least one attribute
    q = ('SELECT DISTINCT sub_id, rt_id, restype_name, rt_attributes.attribute'
         ' FROM subs_resources'
         ' JOIN resource_types USING (rt_id)'
         ' JOIN resource_classes USING (class_id)'
         ' JOIN services USING (sc_id)'
         ' JOIN rt_attributes USING (rt_id)'
         ' JOIN host_attributes USING (host_id)'
         ' WHERE rt_attributes.attribute = host_attributes.host_attribute'
         ' AND services.name IN %s'
         ' AND sub_limit != 0'
         ' AND subs_resources.sub_id IN %s')

    con = uSysDB.connect()
    cur = con.cursor()

    cur.execute(q, (tuple(service_names), tuple(subscription_ids)))
    rt_with_attrs = [
        (sub_id, rt_id, restype_name, rt_attr) for [sub_id, rt_id, restype_name, rt_attr] in cur.fetchall()
        ]

    # resource w/o attribute
    q = ('SELECT DISTINCT sub_id, rt_id, restype_name'
         ' FROM subs_resources'
         ' JOIN resource_types USING (rt_id)'
         ' JOIN resource_classes USING (class_id)'
         ' JOIN services USING (sc_id)'
         ' LEFT JOIN rt_attributes USING (rt_id)'
         ' WHERE services.name IN %s'
         ' AND sub_limit != 0'
         ' AND subs_resources.sub_id IN %s'
         ' GROUP BY sub_id, rt_id, restype_name'
         ' HAVING COUNT(rt_attributes.rt_id) = 0')

    cur.execute(q, (tuple(service_names), tuple(subscription_ids)))
    rt_no_attrs = [(sub_id, rt_id, restype_name, None) for [sub_id, rt_id, restype_name] in cur.fetchall()]

    return set(rt_with_attrs + rt_no_attrs)
